import os
import glob
from yggdrasil import components, constants
from yggdrasil.drivers.CompiledModelDriver import (
    CompiledModelDriver, CompilerBase)


class BuildToolBase(CompilerBase):
    r"""Base class for build tools."""

    is_build_tool = True
    build_language = None

    @staticmethod
    def before_registration(cls):
        r"""Operations that should be performed to modify class attributes prior
        to registration including things like platform dependent properties and
        checking environment variables for default settings.
        """
        CompilerBase.before_registration(cls)
        if cls.build_language is None:
            cls.build_language = cls.toolname

    @classmethod
    def get_tool_suffix(cls):
        r"""Get the string that should be added to tool products based on the
        tool used.

        Returns:
            str: Suffix that should be added to tool products to indicate the
                tool used.

        """
        return ""

    @classmethod
    def get_flags(cls, **kwargs):
        r"""Get compilation flags, replacing outfile with target.

        Args:
            **kwargs: Additional keyword arguments are passed to the parent
                class's method.

        Returns:
            list: Compiler flags.

        """
        kwargs.pop('target_compiler', None)
        kwargs.pop('target_linker', None)
        return super(BuildToolBase, cls).get_flags(**kwargs)


class BuildModelDriver(CompiledModelDriver):
    r"""Class for running build file compiled drivers.

    Args:
        name (str): Driver name.
        args (str, list): Executable that should be created (make target) and
            any arguments for the executable.
        target (str, optional): Make target that should be built to create the
            model executable. Defaults to None.
        target_language (str, optional): Language that the target is written in.
            Defaults to None and will be set based on the source files provided.
        target_compiler (str, optional): Compilation tool that should be used
            to compile the target language. Defaults to None and will be set
            based on the selected language driver.
        target_linker (str, optional): Compilation tool that should be used
            to link the target language. Defaults to None and will be set
            based on the selected language driver.
        target_compiler_flags (list, optional): Compilation flags that should
            be passed to the target language compiler. Defaults to [].
        target_linker_flags (list, optional): Linking flags that should
            be passed to the target language linker. Defaults to [].
        env_compiler (str, optional): Environment variable where the compiler
            executable should be stored for use within the Makefile. If not
            provided, this will be determined by the target language driver.
        env_compiler_flags (str, optional): Environment variable where the
            compiler flags should be stored (including those required to compile
            against the |yggdrasil| interface). If not provided, this will
            be determined by the target language driver.
        env_linker (str, optional): Environment variable where the linker
            executable should be stored for use within the Makefile. If not
            provided, this will be determined by the target language driver.
        env_linker_flags (str, optional): Environment variable where the
            linker flags should be stored (including those required to link
            against the |yggdrasil| interface). If not provided, this will
            be determined by the target language driver.
        **kwargs: Additional keyword arguments are passed to parent class.

    Attributes:
        buildfile (str): Path to file containing build instructions.
        builddir (str): Path to directory where build products will be saved.
        sourcedir (str): Path to directory where source files are located.
        compiledir (str): Path to directory where compilation call should be
            made from. Defaults to working_dir.
        target (str): Name of executable that should be created and called.
        target_language (str): Language that the target is written in.
        target_language_driver (ModelDriver): Language driver for the target
            language.
        target_compiler (str): Compilation tool that should be used to
            compile the target language.
        target_linker (str): Compilation tool that should be used to
            link the target language.
        target_compiler_flags (list): Compilation flags that should be
            passed to the target language compiler.
        target_linker_flags (list): Linking flags that should be passed
            to the target language linker.
        env_compiler (str): Compiler environment variable.
        env_compiler_flags (str): Compiler flags environment variable.
        env_linker (str): Linker environment variable.
        env_linker_flags (str): Linker flags environment variable.

    Class Attributes:
        built_where_called (bool): If True, it is assumed that compilation
            output will be saved in the same directory from which the
            compilation command is issued.

    Raises:
        RuntimeError: If neither the IPC or ZMQ C libraries are available.

    """
    _schema_properties = {
        'buildfile': {'type': 'string'},
        'builddir': {'type': 'string'},
        'target': {'type': 'string'},
        'target_language': {'type': 'string'},
        'target_compiler': {'type': 'string'},
        'target_linker': {'type': 'string'},
        'target_compiler_flags': {'type': 'array',
                                  'items': {'type': 'string'}},
        'target_linker_flags': {'type': 'array',
                                'items': {'type': 'string'}},
        'env_compiler': {'type': 'string'},
        'env_compiler_flags': {'type': 'string'},
        'env_linker': {'type': 'string'},
        'env_linker_flags': {'type': 'string'}}
    executable_type = 'build'
    base_languages = ['c']
    built_where_called = False
    sourcedir_as_sourcefile = False
    full_language = False
    is_build_tool = True
    buildfile_base = None

    def __init__(self, *args, **kwargs):
        self.target_language_driver = None
        self._target_language_info = None
        super(BuildModelDriver, self).__init__(*args, **kwargs)

    @staticmethod
    def after_registration(cls, **kwargs):
        r"""Operations that should be performed to modify class attributes after
        registration. For compiled languages this includes selecting the
        default compiler. The order of precedence is the config file 'compiler'
        option for the language, followed by the environment variable set by
        _compiler_env, followed by the existing class attribute.
        """
        for k in ['linker', 'archiver']:
            if k in cls._config_keys:
                cls._config_keys.remove(k)
        if getattr(cls, 'default_compiler', None) is None:
            cls.default_compiler = cls.language
        CompiledModelDriver.after_registration(cls, **kwargs)
        
    def parse_arguments(self, args, **kwargs):
        r"""Sort arguments based on their syntax to determine if an argument
        is a source file, compilation flag, or runtime option/flag that should
        be passed to the model executable.

        Args:
            args (list): List of arguments provided.
            **kwargs: Additional keyword arguments are passed to the parent
                class's method.

        """
        # Set builddir before passing to parent class so that builddir is used
        # to normalize the model file path rather than the working directory
        # which may be different.
        default_attr = [('target_language_driver', None),
                        ('target_compiler', None),
                        ('buildfile', None),
                        ('builddir', None),
                        ('sourcedir', None),
                        ('compile_working_dir', None),
                        ('builddir_base', '.'),
                        ('buildfile_base', None)]
        for k, v in default_attr:
            if not hasattr(self, k):
                setattr(self, k, v)
        # Directory that compilation should be called from
        if ((isinstance(self.compile_working_dir, str)
             and not os.path.isabs(self.compile_working_dir))):
            self.compile_working_dir = os.path.realpath(
                os.path.join(self.working_dir, self.compile_working_dir))
        # Source directory
        if self.sourcedir is None:
            self.sourcedir = os.path.dirname(args[0])
        if not os.path.isabs(self.sourcedir):
            self.sourcedir = os.path.normpath(
                os.path.realpath(os.path.join(self.working_dir,
                                              self.sourcedir)))
        if self.sourcedir_as_sourcefile:
            self.source_files = [self.sourcedir]
        # Build file
        if self.buildfile is None:
            self.buildfile = self.buildfile_base
        if not os.path.isabs(self.buildfile):
            for x in set([self.working_dir, self.sourcedir,
                          self.builddir, self.compile_working_dir]):
                if x is not None:
                    y = os.path.normpath(os.path.join(x, self.buildfile))
                    if os.path.isfile(y):
                        self.buildfile = y
                        break
        # Compilation directory
        if self.compile_working_dir is None:
            self.compile_working_dir = os.path.dirname(self.buildfile)
        if not os.path.isabs(self.compile_working_dir):
            self.compile_working_dir = os.path.realpath(
                os.path.join(self.working_dir, self.compile_working_dir))
        # Build directory
        if self.builddir is None:
            if self.built_where_called:
                self.builddir = self.compile_working_dir
            else:
                self.builddir = os.path.join(self.sourcedir, self.builddir_base)
        if not os.path.isabs(self.builddir):
            self.builddir = os.path.realpath(os.path.join(self.working_dir,
                                                          self.builddir))
        kwargs.setdefault('default_model_dir', self.builddir)
        super(BuildModelDriver, self).parse_arguments(args, **kwargs)

    @classmethod
    def get_source_dir(cls, fname=None, source_dir=None):
        if source_dir is None:
            if isinstance(fname, str):
                if os.path.isdir(fname):
                    source_dir = fname
                else:
                    source_dir = os.path.dirname(fname)
            else:  # pragma: debug
                raise RuntimeError("No source file/dir provided")
        return source_dir

    @classmethod
    def get_language_for_buildfile(cls, buildfile, target=None):  # pragma: debug
        r"""Determine the target language based on the contents of a build
        file.

        Args:
            buildfile (str): Full path to the build configuration file.
            target (str, optional): Target that will be built. Defaults to None
                and the default target in the build file will be used.

        """
        raise ValueError("Could not determine source from the buildfile")

    @classmethod
    def get_language_for_source(cls, fname=None, buildfile=None, languages=None,
                                early_exit=False, call_base=False, **kwargs):
                                
        r"""Determine the language that can be used with the provided source
        file(s). If more than one language applies to a set of multiple files,
        the language that applies to the most files is returned.

        Args:
            fname (str, list): The full path to one or more files. If more than
                one is provided, they are iterated over.
            buildfile (str, optional): Full path to the build configuration file.
                Defaults to None and will be searched for.
            languages (list, optional): The list of languages that are acceptable.
                Defaults to None and any language will be acceptable.
            early_exit (bool, optional): If True, the first language identified
                will be returned if fname is a list of files. Defaults to False.
            source_dir (str, optional): Full path to the directory containing
                the source files. Defaults to None and is determiend from
                fname.
            buildfile (str, optional): Full path to the build configuration file.
                Defaults to None and will be searched for.
            target (str, optional): The build target. Defaults to None.
            call_base (bool, optional): If True, the base class's method is
                called directly. Defaults to False.
            **kwargs: Additional keyword arguments are passed to the parent
                class's method.

        Returns:
            str: The language that can operate on the specified file.

        """
        if not (call_base or isinstance(fname, list)):
            source_dir = cls.get_source_dir(
                fname, source_dir=kwargs.get('source_dir', None))
            if source_dir == fname:
                fname = None
            if (buildfile is None) and cls.buildfile_base:
                buildfile = os.path.join(source_dir, cls.buildfile_base)
            if isinstance(buildfile, str) and os.path.isfile(buildfile):
                try:
                    return cls.get_language_for_buildfile(
                        buildfile, target=kwargs.get('target', None))
                except ValueError:  # pragma: debug
                    pass
            try_list = sorted(list(glob.glob(os.path.join(source_dir, '*'))))
            if fname is not None:
                try_list = [fname, try_list]
                early_exit = True
            call_base = True
        else:
            try_list = fname
        if languages is None:
            languages = constants.LANGUAGES['compiled']
        return super(BuildModelDriver, cls).get_language_for_source(
            try_list, early_exit=early_exit, buildfile=buildfile,
            languages=languages, call_base=call_base, **kwargs)

    def set_target_language(self):
        r"""Set the language of the target being compiled (usually the same
        as the language associated with this driver.

        Returns:
            str: Name of language.

        """
        if self.target_language is None:
            try:
                self.target_language = self.get_language_for_source(
                    fname=self.model_src, source_dir=self.sourcedir,
                    buildfile=self.buildfile, target=self.target)
            except ValueError:
                self.target_language = 'c'
        if self.target_language is not None:
            if self.target_language_driver is None:
                self.target_language_driver = components.import_component(
                    'model', self.target_language)
            if self.target_compiler is None:
                self.target_compiler = self.target_language_driver.get_tool(
                    'compiler', return_prop='name')
        return self.target_language

    @classmethod
    def get_target_language_info(cls, target_language_driver=None,
                                 target_language='c',
                                 target_compiler=None, target_compiler_flags=None,
                                 target_linker=None, target_linker_flags=None,
                                 logging_level=None, without_wrapper=False,
                                 compiler_flag_kwargs=None,
                                 linker_flag_kwargs=None, **kwargs):
        r"""Get a dictionary of information about language compilation tools.

        Args:
            target_language_driver (ModelDriver, optional): Driver associated
                with the target language. If not provided, one will be created
                based on 'target_language'.
            target_language (str, optional): Language to get info for.
                Defaults to 'c'.
            target_compiler (str, optional): Compilation tool that should be
                used to compile the target language. Defaults to None and will
                be set based on the selected language driver.
            target_linker (str, optional): Compilation tool that should be
                used to link the target language. Defaults to None and will
                be set based on the selected language driver.
            target_compiler_flags (list, optional): Compilation flags that
                should be passed to the target language compiler. Defaults
                to [].
            target_linker_flags (list, optional): Linking flags that should
                be passed to the target language linker. Defaults to [].
            logging_level (int, optional): The numeric logging level that
                should be passed as a definition. Defaults to None and is
                ignored.
            without_wrapper (bool, optional): If True, the returned info
                will be updated for compilation when a build wrapper was
                not created. Defaults to False.
            compiler_flag_kwargs (dict, optional): Keyword arguments to pass
                to the get_compiler_flags method. Defaults to None.
            linker_flag_kwargs (dict, optional): Keyword arguments to pass
                to the get_linker_flags method. Defaults to None.
            **kwargs: Additional keyword arguments are added to the output
                dictionary.

        Returns:
            dict: Information about language compilers and linkers.

        """
        if target_language_driver is None:
            target_language_driver = components.import_component(
                'model', target_language)
        compiler = target_language_driver.get_tool(
            'compiler', toolname=target_compiler)
        if target_linker is None:
            linker = compiler.linker()
        else:
            linker = target_language_driver.get_tool(
                'linker', toolname=target_linker)
        if compiler_flag_kwargs is None:
            compiler_flag_kwargs = {}
        if linker_flag_kwargs is None:
            linker_flag_kwargs = {}
        out = {
            'driver': target_language_driver,
            'compiler': compiler,
            'compiler_executable': compiler.get_executable(full_path=True),
            'compiler_env': compiler.default_executable_env,
            'compiler_flags_env': compiler.default_flags_env,
            'linker': linker,
            'linker_executable': linker.get_executable(full_path=True),
            'linker_env': linker.default_executable_env,
            'linker_flags_env': linker.default_flags_env,
            'env': {},
        }
        if ((out['compiler_flags_env']
             and (not isinstance(out['compiler_flags_env'], list)))):
            out['compiler_flags_env'] = [out['compiler_flags_env']]
        if ((out['linker_flags_env']
             and (not isinstance(out['linker_flags_env'], list)))):
            out['linker_flags_env'] = [out['linker_flags_env']]
        default_compiler_kws = dict(
            toolname=compiler.toolname, skip_defaults=True,
            flags=target_compiler_flags,
            dont_skip_env_defaults=True, for_model=True, dry_run=True,
            dont_link=True, logging_level=logging_level)
        default_linker_kws = dict(
            toolname=linker.toolname, skip_defaults=True,
            flags=target_linker_flags,
            dont_skip_env_defaults=True, for_model=True, dry_run=True)
        for k, v in default_compiler_kws.items():
            compiler_flag_kwargs.setdefault(k, v)
        for k, v in default_linker_kws.items():
            linker_flag_kwargs.setdefault(k, v)
        out.update(
            compiler_flags=target_language_driver.get_compiler_flags(
                **compiler_flag_kwargs),
            linker_flags=target_language_driver.get_linker_flags(
                **linker_flag_kwargs))
        # yggdrasil requires that linking be done in C++
        if (((compiler.languages[0].lower() == 'c')
             and ('-lstdc++' not in out['linker_flags']))):
            out['linker_flags'].append('-lstdc++')
        out.update(**kwargs)
        return out
        
    @property
    def target_language_info(self):
        r"""dict: Information about the underlying language."""
        if self._target_language_info is None:
            kws = dict(
                target_language_driver=self.target_language_driver,
                target_compiler=self.target_compiler,
                target_compiler_flags=self.target_compiler_flags,
                target_linker=self.target_linker,
                target_linker_flags=self.target_linker_flags,
                logging_level=self.numeric_logging_level)
            for x in ['compiler', 'linker']:
                if getattr(self, 'env_%s' % x):
                    kws['%s_env' % x] = (
                        getattr(self, 'env_%s' % x))
                if getattr(self, 'env_%s_flags' % x):
                    kws['%s_flags_env' % x] = (
                        getattr(self, 'env_%s_flags' % x))
            self._target_language_info = self.get_target_language_info(**kws)
        return self._target_language_info
        
    @classmethod
    def is_source_file(cls, fname):
        r"""Determine if the provided file name points to a source files for
        the associated programming language by checking the extension.

        Args:
            fname (str): Path to file.

        Returns:
            bool: True if the provided file is a source file, False otherwise.

        """
        for lang in constants.LANGUAGES['compiled']:
            drv = components.import_component('model', lang)
            if drv.is_source_file(fname):
                return True
        return False

    def compile_dependencies_instance(self, *args, **kwargs):
        r"""Compile dependencies specifically for this instance."""
        if (((self.target_language_driver is not None)
             and (not kwargs.get('dry_run', False)))):
            self.target_language_driver.compile_dependencies(
                toolname=self.target_compiler)
        
    def compile_model(self, **kwargs):
        r"""Compile model executable(s).

        Args:
            **kwargs: Keyword arguments are passed on to the parent class's
                method.

        """
        kwargs['working_dir'] = self.compile_working_dir
        kwargs['target_compiler'] = self.target_compiler
        return super(BuildModelDriver, self).compile_model(**kwargs)
        
    def cleanup(self):
        r"""Remove compiled executable."""
        if (self.model_file is not None) and os.path.isfile(self.model_file):
            self.compile_model(target='clean')
        super(BuildModelDriver, self).cleanup()

    @classmethod
    def call_compiler(cls, src, **kwargs):
        r"""Compile a source file into an executable or linkable object file,
        checking for errors.

        Args:
            src (str): Full path to source file.
            **kwargs: Additional keyword arguments are passed to the parent
                class's method.

        Returns:
            str: Full path to compiled source.

        """
        if ('env' not in kwargs) and (not kwargs.get('dry_run', False)):
            kws = {}
            for k in ['target_language_driver', 'target_language',
                      'target_compiler', 'target_compiler_flags',
                      'target_linker', 'target_linker_flags', 'logging_level']:
                if k in kwargs:
                    kws[k] = kwargs[k]
            if not (('target_language_driver' in kws) or ('target_language' in kws)):
                if src:
                    kws['target_language'] = cls.get_language_for_source(src)
            language_info = cls.get_target_language_info(
                without_wrapper=True, **kws)
            kwargs['env'] = cls.set_env_compiler(language_info=language_info)
            if not kwargs.get('target_compiler', None):
                kwargs['target_compiler'] = language_info['compiler'].toolname
        return super(BuildModelDriver, cls).call_compiler(src, **kwargs)

    @classmethod
    def fix_path(cls, path, for_env=False):
        r"""Update a path.

        Args:
            path (str): Path that should be formatted.
            for_env (bool, optional): If True, the path is formatted for use in
                and environment variable. Defaults to False.

        Returns:
            str: Updated path.

        """
        return path
        
    @classmethod
    def set_env_compiler(cls, language_info=None, **kwargs):
        r"""Get environment variables that should be set for the compilation
        process.

        Args:
            language_info (dict): Language compilation tool information.
            **kwargs: Additional keyword arguments are passed to the parent
                class's method.

        Returns:
            dict: Environment variables for the model process.

        """
        assert(language_info is not None)
        kwargs['compiler'] = language_info['compiler']
        out = super(BuildModelDriver, cls).set_env_compiler(**kwargs)
        if language_info['compiler_env'] and language_info['compiler_executable']:
            out[language_info['compiler_env']] = cls.fix_path(
                language_info['compiler_executable'], for_env=True)
        if language_info['linker_env'] and language_info['linker_executable']:
            out[language_info['linker_env']] = cls.fix_path(
                language_info['linker_executable'], for_env=True)
        if language_info['compiler_flags_env']:
            for x in language_info['compiler_flags_env']:
                out[x] = ' '.join(language_info['compiler_flags'])
        if language_info['linker_flags_env']:
            for x in language_info['linker_flags_env']:
                out[x] = ' '.join(language_info['linker_flags'])
        out.update(language_info['env'])
        return out
    
    def set_env(self, **kwargs):
        r"""Get environment variables that should be set for the model process.

        Args:
            **kwargs: Additional keyword arguments are passed to the parent
                class's method.

        Returns:
            dict: Environment variables for the model process.

        """
        if kwargs.get('for_compile', False) and (self.target_language is not None):
            kwargs.setdefault('compile_kwargs', {})
            kwargs['compile_kwargs']['language_info'] = self.target_language_info
        out = super(BuildModelDriver, self).set_env(**kwargs)
        if not kwargs.get('for_compile', False):
            kwargs['existing'] = out
            if hasattr(self.target_language_driver, 'set_env_class'):
                out = self.target_language_driver.set_env_class(**kwargs)
        return out



import re
import json
import numpy as np
from yggdrasil import tools, units
from yggdrasil.serialize.AsciiMapSerialize import AsciiMapSerialize
from yggdrasil.metaschema.encoder import JSONReadableEncoder


class WOFOSTParamSerialize(AsciiMapSerialize):
    r"""Class for serializing/deserializing WOFOST parameter files."""

    _seritype = 'wofost'
    _schema_subtype_description = ('Serialization of mapping between '
                                   'keys and scalar or array values '
                                   'as used in the WOFOST parameter files.')
    _schema_properties = {
        'delimiter': {'type': 'string',
                      'default': ' = '}}
    default_datatype = {'type': 'object'}
    concats_as_str = True
    _delimiter = ' = '
    _array_fmt = '%5.5f, %5.5f'

    def func_serialize(self, args):
        r"""Serialize a message.

        Args:
            args (dict): Python dictionary to be serialized.

        Returns:
            bytes, str: Serialized message.

        """
        lines = []
        for k in args.keys():
            v = args[k]
            if not isinstance(k, (str, bytes)):  # pragma: debug
                raise ValueError("Serialization of non-string keys not supported.")
            iline = tools.bytes2str(k) + self.delimiter
            if isinstance(v, list):
                indent = ' ' * len(iline)
                arr_lines = []
                assert(len(v) == 2)
                assert(v[0].shape == v[1].shape)
                for i in range(len(v[0])):
                    arr_lines.append(self._array_fmt % (v[0][i], v[1][i]))
                iline += (',\n' + indent).join(arr_lines)
            elif isinstance(v, str):
                iline += "\'%s\'" % v
            else:
                iline += json.dumps(v, cls=JSONReadableEncoder)
            lines.append(iline)
        return tools.str2bytes('\n'.join(lines))

    @classmethod
    def parse_units(cls, x):
        r"""Parse units.

        Args:
            x (str): Unit string.

        Returns:
            str: Propertly formatted units.

        """
        replacements = {"kg N": "kg",
                        "kg P": "kg",
                        "kg-1 dry biomass": "kg-1",
                        "kg CH2O": "kg",
                        "days": "d",
                        "cel": "degC"}
        for k, v in replacements.items():
            x = x.replace(k, v)
        if x == '-':
            x = ""
        out = units.convert_R_unit_string(x)
        return out
            
    def func_deserialize(self, msg):
        r"""Deserialize a message.

        Args:
            msg (bytes): Message to be deserialized.

        Returns:
            dict: Deserialized Python dictionary.

        """
        regex = (r'(?:(?:(?P<name>[^\!]*?)\s*=\s*)|(?:\s+))'
                 r'(?P<value1>(?:-?[\d.]+(?:\.[\d+])?)|(?:\'[^\']*?\'))?'
                 r'\s*(?:(?P<unwrapped_units>[^\!\,\']*?))?'
                 r'(?:\,\s+(?P<value2>[^\!\,]*?)'
                 r'(?:\,?))?\s*(?:\!\s*(?P<description>[^\[\]]*?)?\s*'
                 r'(?:\[\s*(?P<units>[^\]]*?)\s*\][^\[\]]*)?)?\s*')
        out = dict()
        lines = tools.bytes2str(msg.split(self.newline), recurse=True)
        k = None
        desc = ''
        k_units = ''
        is_arr = False

        def finalize_element():
            if is_arr:
                out[k][0] = np.array(out[k][0])
                out[k][1] = np.array(out[k][1])
            if k_units:
                if is_arr:
                    if ';' in k_units:
                        u1, u2 = k_units.split(';')
                        u1 = self.parse_units(u1.strip())
                        u2 = self.parse_units(u2.strip())
                        out[k][0] = units.add_units(out[k][0], u1)
                        out[k][1] = units.add_units(out[k][1], u2)
                    else:
                        out[k][1] = units.add_units(
                            out[k][1], self.parse_units(k_units))
                else:
                    out[k] = units.add_units(
                        out[k], self.parse_units(k_units))
        
        for line in lines:
            if (not line.strip()) or line.startswith('**'):
                continue
            match = re.fullmatch(regex, line)
            if not match:  # pragma: debug
                raise Exception("Failed to parse line: '%s'" % line)
            match = match.groupdict()
            if match.get('name', None):
                finalize_element()
                k = match['name']
                if isinstance(match.get('value1', None), str):
                    match['value1'] = match['value1'].strip()
                if isinstance(match.get('value2', None), str):
                    match['value2'] = match['value2'].strip()
                is_arr = bool(match.get('value2', None))
                desc = ''
                k_units = ''
                if is_arr:
                    out[k] = [[], []]
            if match.get('description', ''):
                desc += match['description']
            if match.get('units', ''):
                k_units += match['units']
            if match['value1']:
                if match['value1'].startswith('\'') and match['value1'].endswith('\''):
                    v1 = match['value1'].strip('\'')
                else:
                    try:
                        v1 = json.loads(match['value1'])
                    except json.decoder.JSONDecodeError:
                        if match['value1'].endswith('.'):
                            v1 = json.loads(match['value1'] + '0')
                        else:  # pragma: debug
                            raise
                if is_arr:
                    v2 = json.loads(match['value2'])
                    out[k][0].append(v1)
                    out[k][1].append(v2)
                else:
                    out[k] = v1
        if out:
            finalize_element()
        return out

    @classmethod
    def concatenate(cls, objects, **kwargs):
        r"""Concatenate objects to get object that would be recieved if
        the concatenated serialization were deserialized.

        Args:
            objects (list): Objects to be concatenated.
            **kwargs: Additional keyword arguments are ignored.

        Returns:
            list: Set of objects that results from concatenating those provided.

        """
        total = dict()
        for x in objects:
            total.update(x)
        return [total]
        
    @classmethod
    def get_testing_options(cls):
        r"""Method to return a dictionary of testing options for this class.

        Returns:
            dict: Dictionary of variables to use for testing.

        """
        out = super(WOFOSTParamSerialize, cls).get_testing_options()
        out['exact_contents'] = False
        out['objects'] = [{'CRPNAM': 'Grain maize CSA practicals',
                           'TBASEM': units.add_units(4.0, 'degC'),
                           'TEFFMX': units.add_units(30.0, 'degC'),
                           'TSUMEM': units.add_units(110.0, 'degC*d'),
                           'IDSL': 0,
                           'DLO': units.add_units(-99.0, 'hr'),
                           'DLC': units.add_units(-99.0, 'hr'),
                           'TSUM1': units.add_units(900.0, 'degC*d'),
                           'TSUM2': units.add_units(800.0, 'degC*d'),
                           'DTSMTB': [
                               units.add_units(
                                   np.array([0.0, 10.0, 30.0, 35.0]), 'degC'),
                               units.add_units(
                                   np.array([0.0, 0.0, 24.0, 24.0]), 'degC*d')],
                           'DVSI': 0.0,
                           'DVSEND': 2.0,
                           'SSATB': [
                               units.add_units(
                                   np.array([0.0, 2.0]), 'ha/kg'),
                               units.add_units(
                                   np.array([0.0, 0.0]), 'ha/kg')]}]
        out['empty'] = dict()
        out['contents'] = (
            b'CRPNAM=\'Grain maize CSA practicals\'\n\n'
            b'** emergence\n'
            b'TBASEM   =   4.0    ! lower threshold temp. for emergence [cel]\n'
            b'TEFFMX   =  30.0    ! max. eff. temp. for emergence [cel]\n'
            b'TSUMEM   = 110.     ! temperature sum from sowing to emergence [cel d]\n\n'
            b'** phenology\n'
            b'IDSL     =   0      ! indicates whether pre-anthesis development depends\n'
            b'                    ! on temp. (=0), daylength (=1) , or both (=2)\n'
            b'DLO      = -99.0    ! optimum daylength for development [hr]\n'
            b'DLC      = -99.0    ! critical daylength (lower threshold) [hr]\n'
            b'TSUM1    = 900.     ! temperature sum from emergence to anthesis [cel d]\n'
            b'TSUM2    = 800.     ! temperature sum from anthesis to maturity [cel d]\n'
            b'DTSMTB   =   0.00,    0.00,     ! daily increase in temp. sum\n'
            b'            10.00,    0.00,     ! as function of av. temp. '
            b'[cel; cel d]\n'
            b'            30.00,   24.00,\n'
            b'            35.00,   24.00\n'
            b'DVSI = 0.           ! initial DVS\n'
            b'DVSEND   =   2.00   ! development stage at harvest (= 2.0 at '
            b'maturity [-]))\n'
            b'SSATB =  0.0, 0.0,  ! specific stem area [ha kg-1]\n'
            b'         2.0, 0.0   ! as function of DVS\n')
        return out



import unittest
from yggdrasil.schema import get_schema
from yggdrasil.tests import assert_raises, assert_equal
import yggdrasil.drivers.tests.test_ConnectionDriver as parent
from yggdrasil.drivers.tests.test_ConnectionDriver import (
    _default_comm, _zmq_installed, _ipc_installed, _rmq_installed)


class TestRPCRequestParam(parent.TestConnectionParam):
    r"""Test parameters for RPCRequestDriver class."""

    def __init__(self, *args, **kwargs):
        super(TestRPCRequestParam, self).__init__(*args, **kwargs)
        self.driver = 'RPCRequestDriver'
        self.args = None
        self.attr_list += ['response_drivers', 'clients']
        # Increased to allow forwarding between IPC comms on MacOS
        self.timeout = 5.0
        self.route_timeout = 2 * self.timeout
        # if tools.get_default_comm() == "IPCComm":
        #     self.route_timeout = 120.0
        # self.debug_flag = True
        # self.sleeptime = 0.5
        # self.timeout = 10.0
            
    @property
    def send_comm_kwargs(self):
        r"""dict: Keyword arguments for send comm."""
        out = self.instance.icomm.opp_comm_kwargs()
        out['request_commtype'] = out['commtype']
        out['commtype'] = 'client'
        return out

    @property
    def recv_comm_kwargs(self):
        r"""dict: Keyword arguments for recv comm."""
        out = self.instance.ocomm.opp_comm_kwargs()
        out['request_commtype'] = out['commtype']
        out['commtype'] = 'server'
        return out

    
class TestRPCRequestDriverNoStart(TestRPCRequestParam,
                                  parent.TestConnectionDriverNoStart):
    r"""Test class for RPCRequestDriver class without start."""
    
    def test_error_attributes(self):
        r"""Test error raised when trying to access attributes set on recv."""
        err_attr = ['request_id', 'response_address']
        for k in err_attr:
            assert_raises(AttributeError, getattr, self.instance, k)


class TestRPCRequestDriverNoInit(TestRPCRequestParam,
                                 parent.TestConnectionDriverNoInit):
    r"""Test class for RPCRequestDriver class without init."""
    pass
            

class TestRPCRequestDriver(TestRPCRequestParam,
                           parent.TestConnectionDriver):
    r"""Test class for RPCRequestDriver class."""

    def test_send_recv(self, msg_send=None):
        r"""Test routing of a short message between client and server."""
        try:
            if msg_send is None:
                msg_send = self.test_msg
            T = self.instance.start_timeout()
            while ((not T.is_out) and (not self.instance.is_valid)):
                self.instance.sleep()  # pragma: debug
            self.instance.stop_timeout()
            # Send a message to local output
            flag = self.send_comm.send(msg_send)
            assert(flag)
            # Receive on server side, then send back
            flag, srv_msg = self.recv_comm.recv(timeout=self.route_timeout)
            assert(flag)
            assert_equal(srv_msg, msg_send)
            self.instance.printStatus()
            flag = self.recv_comm.send(srv_msg)
            assert(flag)
            # Receive response on client side
            flag, cli_msg = self.send_comm.recv(timeout=self.route_timeout)
            assert(flag)
            assert_equal(cli_msg, msg_send)
        except BaseException:  # pragma: debug
            self.send_comm.printStatus()
            self.instance.printStatus(verbose=True)
            self.recv_comm.printStatus()
            raise

    def test_send_recv_nolimit(self):
        r"""Test routing of a large message between client and server."""
        self.test_send_recv(msg_send=self.msg_long)


# Dynamically create tests based on registered comm classes
s = get_schema()
comm_types = list(s['comm'].schema_subtypes.keys())
for k in comm_types:
    if k in [_default_comm, 'ValueComm', 'value',
             'BufferComm', 'buffer']:  # pragma: debug
        continue
    tcls = type('Test%sRPCRequestDriver' % k,
                (TestRPCRequestDriver, ), {'ocomm_name': k,
                                           'icomm_name': k,
                                           'driver': 'RPCRequestDriver',
                                           'args': 'test'})
    # Flags
    flag_func = None
    if k in ['RMQComm', 'RMQAsyncComm', 'rmq', 'rmq_async']:
        flag_func = unittest.skipIf(not _rmq_installed,
                                    "RMQ Server not running")
    elif k in ['ZMQComm', 'zmq']:
        flag_func = unittest.skipIf(not _zmq_installed,
                                    "ZMQ library not installed")
    elif k in ['IPCComm', 'ipc']:
        flag_func = unittest.skipIf(not _ipc_installed,
                                    "IPC library not installed")
    if flag_func is not None:
        tcls = flag_func(tcls)
    # Add class to globals
    globals()[tcls.__name__] = tcls
    del tcls



import os
import unittest
import signal
import uuid
from yggdrasil import runner, tools, platform, import_as_function
from yggdrasil.tests import assert_raises, requires_language
# from yggdrasil.tests import yamls as sc_yamls
from yggdrasil.examples import yamls as ex_yamls


def test_get_runner():
    r"""Use get_runner to start a run."""
    namespace = "test_get_runner_%s" % str(uuid.uuid4)
    cr = runner.get_runner([ex_yamls['hello']['python']],
                           namespace=namespace)
    cr.run()
    cr.sleep()


def test_get_run():
    r"""Use run function to start a run."""
    namespace = "test_run_%s" % str(uuid.uuid4)
    runner.run([ex_yamls['hello']['python']],
               namespace=namespace)
    runner.run([ex_yamls['model_error']['python']],
               namespace=namespace)


def test_run_process_connections():
    r"""Test run with process based connections."""
    namespace = "test_run_%s" % str(uuid.uuid4)
    runner.run([ex_yamls['hello']['python']],
               connection_task_method='process',
               namespace=namespace)


# def test_runner_error():
#     r"""Start a runner for a model with an error."""
#     cr = runner.get_runner([sc_yamls['error']])
#     cr.run()


# Spawning fake Ctrl-C works locally for windows, but causes hang on appveyor
@unittest.skipIf(platform._is_win, "Signal processing not sorted on windows")
def test_Arunner_interrupt():
    r"""Start a runner then stop it with a keyboard interrupt."""
    cr = runner.get_runner([ex_yamls['hello']['python']])
    if platform._is_win:  # pragma: debug
        cr.debug_log()
    cr.loadDrivers()
    cr.startDrivers()
    cr.set_signal_handler()
    tools.kill(os.getpid(), signal.SIGINT)
    tools.kill(os.getpid(), signal.SIGINT)
    cr.reset_signal_handler()
    cr.waitModels()
    cr.closeChannels()
    cr.cleanup()
    if platform._is_win:  # pragma: debug
        cr.reset_log()


def test_runner_terminate():
    r"""Start a runner, then stop it early."""
    cr = runner.get_runner([ex_yamls['hello']['python']])
    cr.loadDrivers()
    cr.startDrivers()
    cr.printStatus()
    cr.terminate()


def test_runner_error():
    r"""Test error on missing yaml."""
    assert_raises(IOError, runner.YggRunner,
                  ['fake_yaml.yml'], 'test_ygg_run')


def test_import_as_function():
    r"""Test import_as_function."""
    yamlfile = ex_yamls['fakeplant']['python']
    fmodel = import_as_function(yamlfile)
    input_args = {}
    for x in fmodel.arguments:
        input_args[x] = 1.0
    fmodel.model_info()
    result = fmodel(**input_args)
    for x in fmodel.returns:
        assert(x in result)
    result = fmodel(*list(input_args.values()))
    for x in fmodel.returns:
        assert(x in result)
    fmodel.stop()
    fmodel.stop()


def test_import_as_function_server():
    r"""Test import_as_function w/ server function."""
    yamlfile = None
    for x in ex_yamls['rpc_lesson3']['python']:
        if 'server' in x:
            yamlfile = x
            break
    assert(yamlfile)
    fmodel = import_as_function(yamlfile)
    input_args = {}
    for x in fmodel.arguments:
        input_args[x] = 'hello'
    fmodel.model_info()
    result = fmodel(**input_args)
    for x in fmodel.returns:
        assert(x in result)
    result = fmodel(*list(input_args.values()))
    for x in fmodel.returns:
        assert(x in result)
    fmodel.stop()
    fmodel.stop()


@requires_language('c')
def test_import_as_function_C():
    r"""Test import_as_function for C."""
    contents = r"""models:
      - name: c_modelA
        language: c
        args: ./src/model_function_modelA.c
        function: model_function
        inputs: inputA
        outputs: outputA"""
    yamlfile = os.path.join(os.path.dirname(ex_yamls['model_function']['c']),
                            'test_import.yml')
    assert(not os.path.isfile(yamlfile))
    with open(yamlfile, 'w') as fd:
        fd.write(contents)
    try:
        fmodel = import_as_function(yamlfile)
        fmodel.model_info()
        input_args = {}
        for x in fmodel.arguments:
            input_args[x] = b'hello'
        result = fmodel(**input_args)
        for x in fmodel.returns:
            assert(x in result)
        result = fmodel(*list(input_args.values()))
        for x in fmodel.returns:
            assert(x in result)
        fmodel.stop()
        fmodel.stop()
    finally:
        os.remove(yamlfile)



from yggdrasil.communication.CommBase import CommBase


class ValueComm(CommBase):

    _commtype = 'value'
    _schema_subtype_description = ('Constant value.')
    _schema_properties = {
        'count': {'type': 'integer', 'default': 1}}
    no_serialization = True

    def __init__(self, *args, **kwargs):
        self._is_open = False
        super(ValueComm, self).__init__(*args, **kwargs)
        assert(self.direction == 'recv')
        self.remaining = self.count

    @classmethod
    def is_installed(cls, language=None):
        r"""Determine if the necessary libraries are installed for this
        communication class.

        Args:
            language (str, optional): Specific language that should be checked
                for compatibility. Defaults to None and all languages supported
                on the current platform will be checked. If set to 'any', the
                result will be True if this comm is installed for any of the
                supported languages.

        Returns:
            bool: Is the comm installed.

        """
        if language == 'python':
            return True
        return False
        
    def open(self):
        r"""Open the connection."""
        super(ValueComm, self).open()
        self._is_open = True

    def _close(self, *args, **kwargs):
        r"""Close the connection."""
        self._is_open = False
        super(ValueComm, self)._close()

    @property
    def is_open(self):
        r"""bool: True if the connection is open."""
        return self._is_open
    
    @property
    def n_msg_recv(self):
        r"""int: The number of incoming messages in the connection."""
        return self.remaining

    def send(self, *args, **kwargs):
        r"""Send a message."""
        raise RuntimeError("Cannot send to a ValueComm.")

    def _recv(self, **kwargs):
        if self.remaining == 0:
            return True, self.eof_msg
        self.remaining -= 1
        return True, self.address

    @classmethod
    def get_testing_options(cls, **kwargs):
        r"""Method to return a dictionary of testing options for this class.

        Args:
            serializer (str, optional): The name of the serializer that should
                be used. If not provided, the _default_serializer class
                attribute will be used.

        Returns:
            dict: Dictionary of variables to use for testing. Key/value pairs:
                kwargs (dict): Keyword arguments for comms tested with the
                    provided content.
                send (list): List of objects to send to test file.
                recv (list): List of objects that will be received from a test
                    file that was sent the messages in 'send'.
                contents (bytes): Bytes contents of test file created by sending
                    the messages in 'send'.

        """
        msg = ['test', 1.0]
        count = 3
        objects = [msg for _ in range(count)]
        out = {'kwargs': {'address': msg, 'count': count},
               'send': objects,
               'msg': msg,
               'objects': objects,
               'recv': objects}
        return out

    def purge(self):
        r"""Purge all messages from the comm."""
        self.remaining = 0



import uuid
from yggdrasil import multitasking
from yggdrasil.tools import ProxyObject
from yggdrasil.components import ComponentBaseUnregistered
from yggdrasil.communication import (
    CommBase, TemporaryCommunicationError)


FLAG_FAILURE = 0
FLAG_SUCCESS = 1
FLAG_TRYAGAIN = 2


class AsyncComm(ProxyObject, ComponentBaseUnregistered):
    r"""Class for handling asynchronous I/O.

    Args:
        name (str): The name of the message queue.
        async_recv_method (str, optional): The method that should be used
            to receive message into the backlog. Defaults to 'recv'.
        async_send_method (str, optional): The method that should be used
            to send message in the backlog. Defaults to 'send_message'.
        async_recv_kwargs (dict, optional): Keyword arguments to pass to calls
            receiving messages into the backlog. Defaults to {}.
        async_send_method (dict, optional): Keyword arguments to pass to calls
            sending message from the backlog. Defaults to {}.
        **kwargs: Additional keyword arguments are passed to CommBase.
        
    Attributes:
        backlog_ready (multitasking.Event): Event set when there is a
            message in the recv backlog.
        
    """

    __slots__ = ['_backlog_buffer', '_backlog_thread',
                 'backlog_ready', '_used_direct', 'close_on_eof_recv',
                 '_backlog_received_eof', '_used', '_closed',
                 'async_recv_method', 'async_send_method',
                 'async_recv_kwargs', 'async_send_kwargs',
                 '_error_registry']
    __overrides__ = ['_input_args', '_input_kwargs']
    _disconnect_attr = ['backlog_ready', '_backlog_thread', '_wrapped']
    _async_kws = ['async_recv_method', 'async_send_method',
                  'async_recv_kwargs', 'async_send_kwargs']

    def __init__(self, wrapped,
                 async_recv_method='recv', async_send_method='send_message',
                 async_recv_kwargs=None, async_send_kwargs=None):
        self._backlog_buffer = []
        self._backlog_thread = None
        self.backlog_ready = multitasking.Event()
        self._used_direct = False
        self.close_on_eof_recv = wrapped.close_on_eof_recv
        self._used = False
        self._closed = False
        self._backlog_received_eof = False
        self._error_registry = {}
        self.async_recv_method = async_recv_method
        self.async_send_method = async_send_method
        if async_recv_kwargs is None:
            async_recv_kwargs = {}
        if async_send_kwargs is None:
            async_send_kwargs = {}
        self.async_recv_kwargs = async_recv_kwargs
        self.async_send_kwargs = async_send_kwargs
        wrapped.close_on_eof_recv = False
        wrapped.is_async = True
        super(AsyncComm, self).__init__(wrapped)
        # Open backlog to match
        if self._wrapped.is_open:
            self.open()
        if self._wrapped.is_interface:  # pragma: debug
            # atexit.register(self.atexit)
            raise RuntimeError("Use of async comm inside model is untested")

    def __reduce__(self):
        rv = list(super(AsyncComm, self).__reduce__())
        (wrapped, ) = rv[1]
        wrapped.close_on_eof_recv = self.close_on_eof_recv
        rv[1] = (wrapped, )
        return tuple(rv)

    def precheck(self, *args, **kwargs):
        CommBase.CommBase.precheck(self, *args, **kwargs)

    def printStatus(self, *args, **kwargs):
        r"""Print status of the communicator."""
        lines = ['%-15s: %s' % ('open (backlog)', self.is_open_backlog),
                 '%-15s: %s' % ('close called (backlog)', self._closed)]
        if self.direction == 'send':
            lines.append(
                '%-15s: %s' % ('nsent (backlog)', self.n_msg_backlog_send))
        else:
            lines.append(
                '%-15s: %s' % ('nrecv (backlog)', self.n_msg_backlog_recv))
        kwargs.setdefault('extra_lines_after', [])
        kwargs['extra_lines_after'] += lines
        self._wrapped.printStatus(*args, **kwargs)
        
    @property
    def backlog_thread(self):
        r"""tools.YggTask: Task that will handle sinding or receiving
        backlogged messages."""
        if self._backlog_thread is None:
            if self.direction == 'send':
                self._backlog_thread = CommBase.CommTaskLoop(
                    self, target=self.run_backlog_send, suffix='SendBacklog')
            else:
                self._backlog_thread = CommBase.CommTaskLoop(
                    self, target=self.run_backlog_recv, suffix='RecvBacklog')
        return self._backlog_thread

    @property
    def errors(self):
        r"""list: Errors raised by the wrapped comm or async thread."""
        out = self._wrapped.errors
        if self._backlog_thread:
            out += self._backlog_thread.errors
            self._backlog_thread.errors = []
        return out

    def atexit(self):  # pragma: debug
        r"""Close operations."""
        if (self.direction == 'send') and self.is_open_backlog:
            self.linger()

    def open(self):
        r"""Open the connection by connecting to the queue."""
        self._wrapped.open()
        self._wrapped.suppress_special_debug = True
        if self._wrapped.is_open and (not self.is_open_backlog):
            self.backlog_thread.start()

    def close(self, linger=False):
        r"""Close the connection.

        Args:
            linger (bool, optional): If True, drain messages before closing the
                comm. Defaults to False.

        """
        with self._closing_thread.lock:
            self._wrapped.close(linger=linger)
            self._close_backlog(wait=linger)
        with self.backlog_thread.lock:
            self._closed = True
        # Requeue messages for other servers
        if ((self.is_server and (self.model_copies > 1)
             and self._backlog_buffer)):  # pragma: debug
            # client_kws = self.opp_comm_kwargs
            # client_kws.update(use_async=False)
            # client = get_comm(self.name + '_client', **client_kws)
            # for msg in self._backlog_buffer:
            #     self.info("Resending: %s", msg)
            #     client.send(msg[0])
            # client.close()
            raise RuntimeError("Returning backlogged messages to the server "
                               "is untested.")

    def _close_backlog(self, wait=False):
        r"""Close the backlog thread."""
        self.debug('')
        if self._backlog_thread is not None:
            self.backlog_thread.set_break_flag()
        self.backlog_ready.set()
        if wait and (self._backlog_thread is not None):  # pragma: intermittent
            self.backlog_thread.wait(key=str(uuid.uuid4()))
        if hasattr(self._wrapped, '_close_backlog'):
            self._wrapped._close_backlog(wait=wait)

    @property
    def is_open(self):
        r"""bool: True if the backlog is open."""
        if self.direction == 'send':
            return self._wrapped.is_open and self.is_open_backlog
        else:
            return ((self.is_open_backlog or (self.n_msg_backlog > 0))
                    and (not self._closed))

    @property
    def is_open_direct(self):
        r"""bool: True if the direct comm is not None."""
        return self._wrapped.is_open

    @property
    def is_open_backlog(self):
        r"""bool: True if the backlog thread is running."""
        return ((self._backlog_thread is not None)
                and (not self.backlog_thread.was_break)
                and (self.backlog_thread.is_alive()))

    @property
    def is_closed(self):
        r"""bool: True if the connection is closed."""
        return (not self.is_open)

    @property
    def n_msg(self):
        r"""int: The number of messages in the connection."""
        if self.direction == 'recv':
            return self.n_msg_recv
        else:
            return self.n_msg_send

    @property
    def n_msg_recv(self):
        r"""int: Number of messages in the receive backlog."""
        return self.n_msg_backlog_recv

    @property
    def n_msg_send(self):
        r"""int: Number of messages in the send backlog."""
        return self.n_msg_backlog_send

    @property
    def n_msg_direct_recv(self):
        r"""int: Number of messages currently being routed in recv."""
        return self._wrapped.n_msg_recv

    @property
    def n_msg_direct_send(self):
        r"""int: Number of messages currently being routed in send."""
        return self._wrapped.n_msg_send

    @property
    def n_msg_direct(self):
        r"""int: Number of messages currently being routed."""
        if self.direction == 'send':
            return self.n_msg_direct_send
        else:
            return self.n_msg_direct_recv

    @property
    def n_msg_backlog_recv(self):
        r"""int: Number of messages in the receive backlog."""
        if self.direction == 'recv':
            return self.n_msg_backlog
        return 0  # pragma: debug

    @property
    def n_msg_backlog_send(self):
        r"""int: Number of messages in the send backlog."""
        if self.direction == 'send':
            return self.n_msg_backlog
        return 0  # pragma: debug

    @property
    def n_msg_backlog(self):
        r"""int: Number of messages in the backlog."""
        if (self.direction == 'recv') or self.is_open_backlog:
            return len(self.backlog_buffer)
        return 0

    @property
    def n_msg_recv_drain(self):
        r"""int: Number of messages in the receive backlog and direct comm."""
        return self.n_msg_direct_recv + self.n_msg_backlog_recv

    @property
    def n_msg_send_drain(self):
        r"""int: Number of messages in the send backlog and direct comm."""
        return self.n_msg_direct_send + self.n_msg_backlog_send

    @property
    def is_confirmed_send(self):
        r"""bool: True if all sent messages have been confirmed."""
        if not self._wrapped.is_confirmed_send:
            return False
        return (self.n_msg_send == 0)
        
    @property
    def backlog_buffer(self):
        r"""list: Messages that have been received."""
        with self.backlog_thread.lock:
            return self._backlog_buffer

    def add_backlog(self, payload):
        r"""Add a message to the backlog of messages.

        Args:
            payload (tuple): Arguments and keyword argumetns for send
                or data from receive.

        """
        with self.backlog_thread.lock:
            self.debug("Added message to %s backlog.",
                       self.direction)
            if not self._closed:
                self._backlog_buffer.append(payload)
                self.backlog_ready.set()

    def pop_backlog(self):
        r"""Pop a message from the front of the backlog.

        Returns:
            tuple: First backlogged send arguments/keyword arguments
                or received data.

        """
        with self.backlog_thread.lock:
            out = self._backlog_buffer.pop(0)
            self.debug("Removed message from backlog.")
            if len(self._backlog_buffer) == 0:
                self.backlog_ready.clear()
        return out

    def run_backlog_send(self):
        r"""Continue trying to send buffered messages."""
        if not self.send_backlog():  # pragma: debug
            self.debug("Stopping because send_backlog failed")
            self._close_backlog()
            return
        self.periodic_debug('run_backlog_send', period=1000)(
            "Sleeping (is_confirmed_send=%s, n_msg_send=%d)",
            str(self.is_confirmed_send), self.n_msg_backlog_send)
        self.sleep()

    def run_backlog_recv(self):
        r"""Continue buffering received messages."""
        if self.backlog_thread.main_terminated:  # pragma: debug
            self.debug("Main thread terminated")
            self.close()
            return
        if not self.recv_backlog():
            self.debug("Failure to receive message into backlog.")
            self._close_backlog()
            return
        self.periodic_debug('run_backlog_recv', period=1000)(
            "Sleeping (is_confirmed_recv=%s)",
            str(self.is_confirmed_recv))
        self.sleep()

    def send_direct(self, *args, **kwargs):
        r"""Send a message directly to the underlying comm."""
        self.periodic_debug("send_direct", period=1000)(
            "Sending message to %s", self.address)
        self.suppress_special_debug = True
        kwargs.update(self.async_send_kwargs)
        try:
            kwargs.setdefault('timeout', 0)
            if getattr(self._wrapped, self.async_send_method)(*args, **kwargs):
                async_flag = FLAG_SUCCESS
                self._used_direct = True
            else:
                async_flag = FLAG_FAILURE
            self._error_registry = {}
        except TemporaryCommunicationError:
            async_flag = FLAG_TRYAGAIN
        self.suppress_special_debug = False
        return async_flag

    def recv_direct(self):
        r"""Receive a message directly from the underlying comm."""
        self.periodic_debug("recv_direct", period=1000)(
            "Receiving message from %s", self.address)
        self.suppress_special_debug = True
        msg = None
        kwargs = self.async_recv_kwargs
        try:
            kwargs.setdefault('timeout', 0)
            if self.async_recv_method != 'recv_message':
                kwargs['return_message_object'] = True
            msg = getattr(self._wrapped, self.async_recv_method)(**kwargs)
            if msg.flag in [CommBase.FLAG_EMPTY, CommBase.FLAG_SKIP]:
                async_flag = FLAG_TRYAGAIN  # pragma: debug
            elif msg.flag in [CommBase.FLAG_SUCCESS, CommBase.FLAG_EOF]:
                async_flag = FLAG_SUCCESS
                self._used_direct = True
                if msg.flag == CommBase.FLAG_EOF:
                    self._backlog_received_eof = True
            elif msg.flag == CommBase.FLAG_FAILURE:
                async_flag = FLAG_FAILURE
            else:  # pragma: debug
                raise Exception("Unsupported flag: %s" % msg.flag)
            self._error_registry = {}
        except TemporaryCommunicationError:
            async_flag = FLAG_TRYAGAIN
        self.suppress_special_debug = False
        return async_flag, msg

    def send_backlog(self):
        r"""Send a message from the send backlog to the queue."""
        if len(self.backlog_buffer) == 0:
            flag = True
        else:
            iargs, ikwargs = self.backlog_buffer[0]
            async_flag = self.send_direct(*iargs, **ikwargs)
            flag = bool(async_flag)
            if async_flag == FLAG_SUCCESS:
                self.pop_backlog()
        self.confirm_send()
        return flag

    def recv_backlog(self):
        r"""Check for any messages in the queue and add them to the recv
        backlog."""
        if not self.is_open_direct:
            flag = False
        elif self._backlog_received_eof and self.close_on_eof_recv:
            # Don't keep receiving, but don't close so that this thread
            # can continue confirmation until the EOF is actually received
            flag = True
        else:
            async_flag, msg = self.recv_direct()
            flag = bool(async_flag)
            if async_flag == FLAG_SUCCESS:
                self.add_backlog(msg)
        self.confirm_recv()
        return flag

    def send_message(self, msg, **kwargs):
        r"""Send a message encapsulated in a CommMessage object.

        Args:
            msg (CommMessage): Message to be sent.
            **kwargs: Additional keyword arguments are passed to _safe_send.

        Returns:
            bool: Success or failure of send.
        
        """
        # This is required so that call to send_message for work comms
        # in CommBase.send_message will put the messages in the backlog
        kwargs['dont_prepare'] = True
        return self.send(msg, **kwargs)
        
    def send(self, *args, dont_prepare=False, **kwargs):
        r"""Send a message to the backlog.

        Args:
            *args: All arguments are assumed to be part of the message.
            **kwargs: All keywords arguments are passed to comm _send method.

        Returns:
            bool: Success or failure of sending the message.

        """
        self.precheck('send')
        kws_prepare = {k: kwargs.pop(k) for k in self._prepare_message_kws
                       if k in kwargs}
        if not self.is_open_direct:  # pragma: debug
            return False
        if dont_prepare:
            assert((len(args) == 1) and isinstance(args[0], CommBase.CommMessage))
            msg = args[0]
        else:
            msg = self._wrapped.prepare_message(*args, **kws_prepare)
        if not self.backlog_ready.is_set():
            async_flag = self.send_direct(msg, **kwargs)
            if async_flag != FLAG_TRYAGAIN:
                return bool(async_flag)
        self.add_backlog(((msg, ), kwargs))
        self._used = True
        return True

    def send_eof(self, *args, **kwargs):
        r"""Send the EOF message as a short message.
        
        Args:
            *args: All arguments are passed to comm send.
            **kwargs: All keywords arguments are passed to comm send.

        Returns:
            bool: Success or failure of send.

        """
        return self.send(self.eof_msg, *args, **kwargs)

    def recv_message(self, *args, **kwargs):
        r"""Receive a message.

        Args:
            *args: Arguments are passed to the response comm's recv_message method.
            **kwargs: Keyword arguments are passed to the response comm's recv_message
                method.

        Returns:
            CommMessage: Received message.

        """
        # This is required so that call to recv_message for work comms
        # in CommBase.recv_message will retrieve messages from the backlog
        kwargs['return_message_object'] = True
        kwargs['dont_finalize'] = True
        return self.recv(*args, **kwargs)
        
    def recv(self, timeout=None, return_message_object=False, dont_finalize=False,
             **kwargs):
        r"""Receive a message.

        Args:
            *args: All arguments are passed to comm _recv method.
            return_message_object (bool, optional): If True, the full wrapped
                CommMessage message object is returned instead of the tuple.
                Defaults to False.
            dont_finalize (bool, optional): If True, finalize_message will not
                be called even if async_recv_method is 'recv_message'. Defaults
                to False.
            **kwargs: All keywords arguments are passed to comm _recv method.

        Returns:
            tuple (bool, obj): Success or failure of receive and received
                message.

        """
        self.precheck('recv')
        # Sleep until there is a message
        if timeout is None:
            timeout = kwargs.get('timeout', self.recv_timeout)
        T = self.start_timeout(timeout, key_suffix='.recv:backlog')
        while (not T.is_out) and (not self.backlog_ready.is_set()):
            self.backlog_ready.wait(self.sleeptime)
        self.stop_timeout(key_suffix='.recv:backlog')
        # Handle absence of messages
        if self.n_msg_backlog == 0:
            self.verbose_debug("No messages waiting.")
            if self.is_closed:
                self.info(("No messages waiting and comm closed."
                           "%s, %s, %s")
                          % (self.backlog_thread is not None,
                             not self.backlog_thread.was_break,
                             self.backlog_thread.is_alive()))
                self.printStatus()
                if self.backlog_thread.was_break:
                    self.info("Break stack:\n%s",
                              self.backlog_thread.break_stack)
                out = CommBase.CommMessage(flag=CommBase.FLAG_FAILURE)
            else:
                out = CommBase.CommMessage(flag=CommBase.FLAG_EMPTY,
                                           args=self.empty_obj_recv)
        # Return backlogged message
        else:
            self.debug('Returning backlogged received message')
            out = self.pop_backlog()
            if not dont_finalize:
                if (out.flag == CommBase.FLAG_EOF) and self.close_on_eof_recv:
                    self.close()
                    out.flag = CommBase.FLAG_FAILURE
                self._used = True
        if not dont_finalize:
            kws_finalize = {k: kwargs.pop(k) for k in self._finalize_message_kws
                            if k in kwargs}
            if self.async_recv_method != 'recv_message':
                out.finalized = False
                kws_finalize['skip_processing'] = True
            out = self._wrapped.finalize_message(out, **kws_finalize)
        if not return_message_object:
            out = (bool(out.flag), out.args)
        return out

    def purge(self):
        r"""Purge all messages from the comm."""
        self._wrapped.purge()
        with self.backlog_thread.lock:
            self.backlog_ready.clear()
            self._backlog_buffer = []

    # ALIASES
    def send_nolimit(self, *args, **kwargs):
        r"""Alias for send_nolimit on wrapped comm."""
        return CommBase.CommBase.send_nolimit(self, *args, **kwargs)

    def recv_nolimit(self, *args, **kwargs):
        r"""Alias for recv_nolimit on wrapped comm."""
        return CommBase.CommBase.recv_nolimit(self, *args, **kwargs)

    def send_array(self, *args, **kwargs):
        r"""Alias for send_array on wrapped comm."""
        return CommBase.CommBase.send_array(self, *args, **kwargs)

    def recv_array(self, *args, **kwargs):
        r"""Alias for recv_array on wrapped comm."""
        return CommBase.CommBase.recv_array(self, *args, **kwargs)

    def send_dict(self, *args, **kwargs):
        r"""Alias for send_dict on wrapped comm."""
        return CommBase.CommBase.send_dict(self, *args, **kwargs)

    def recv_dict(self, *args, **kwargs):
        r"""Alias for recv_dict on wrapped comm."""
        return CommBase.CommBase.recv_dict(self, *args, **kwargs)
    
    def drain_server_signon_messages(self, **kwargs):
        r"""Drain server signon messages. This should only be used
        for testing purposes."""
        pass



import os
import argparse
import pickle

_dir = os.path.dirname(os.path.realpath(__file__))

# Parse command-line arguments
parser = argparse.ArgumentParser("Simulate root growth over time.")
parser.add_argument('tmin', help='Starting time (in days)', type=float)
parser.add_argument('tmax', help='Ending time (in days)', type=float)
parser.add_argument('tstep', help='Time step (in days)', type=float)
args = parser.parse_args()
tmin = args.tmin
tmax = args.tmax
tstep = args.tstep

# Set initial conditions
mass = 0.0
t = tmin
times = []
masses = []

# Check if model is running as a part of an yggdrasil integration
with_yggdrasil = os.environ.get('YGG_SUBPROCESS', False)

# If the model is running as part of an yggdrasil integration, import
# the relevant yggdrasil routines and use the interface routine to
# complete the connection defined in the YAML
if with_yggdrasil:
    from yggdrasil import units
    from yggdrasil.languages.Python.YggInterface import YggTimesync
    shoot2root = YggTimesync('shoot2root')
    
# Continue simulation until time limit is reached
while t <= tmax:

    # If running as part an yggdrasil integration, send the time and
    # mass to the timesync channel and then updated the mass based on
    # the returned state
    if with_yggdrasil:
        root_state = {'mass': units.add_units(mass, 'kg')}
        flag, total_state = shoot2root.call(units.add_units(t, 'days'),
                                            root_state)
        if not flag:
            raise Exception("Error performing time-step synchronization "
                            "with shoot model.")
        
        # Compute the scale factor using total mass, stripping units
        # of the result to allow use with original code
        # (pretend this is a biologically complex calculation)
        scale = units.get_data(
            units.convert_to(
                units.add_units(0.05, 'days-1') * total_state['mass'],
                'kg/day'))

    else:
        # Compute the scale factor
        # (pretend this is a biologically complex calculation)
        scale = 0.2

    # Calculate mass for the time step
    # (pretend this is a biologically complex calculation)
    mass += t * scale

    # Add mass & time to array
    times.append(t)
    masses.append(mass)

    # Advance time step
    t += tstep

# Write the total mass array to output
filename_masses = os.path.join(_dir, '../output/masses.pkl')
with open(filename_masses, 'wb') as fd:
    pickle.dump({'times': times, 'masses': masses}, fd)



"""
This module imports the configuration for yggdrasil.

.. todo::
   Remove reference to environment variables for accessing config options.

"""
import os
import sys
import json
import shutil
import logging
import warnings
import configparser
import copy
import argparse
from contextlib import contextmanager
from collections import OrderedDict
from yggdrasil import tools
env_prefixes = tools.get_env_prefixes()
config_file = '.yggdrasil.cfg'
def_config_file = os.path.join(os.path.dirname(__file__), 'defaults.cfg')
usr_dir = os.path.expanduser('~')
if env_prefixes:
    usr_dir = env_prefixes[-1]
usr_config_file = os.path.join(usr_dir, config_file)
loc_config_file = os.path.join(os.getcwd(), config_file)
logger = logging.getLogger(__name__)


# Set associated environment variables
_cfg_map = {
    ('debug', 'ygg'): {
        'env': 'YGG_DEBUG', 'arg': 'loglevel',
        'help': 'Logging level for yggdrasil operations.'},
    ('debug', 'rmq', 'RMQ_DEBUG'): {
        'env': 'YGG_RMQ_DEBUG', 'arg': 'rmq-loglevel',
        'help': 'Logging level for RabbitMQ operations.'},
    ('debug', 'client'): {
        'env': 'YGG_CLIENT_DEBUG', 'arg': 'client-loglevel',
        'help': 'Logging level for yggdrasil operations on model processes.'},
    ('jsonschema', 'validate_components'): {
        'env': 'YGG_VALIDATE_COMPONENTS', 'arg': 'validate-components',
        'action': 'store_true',
        'help': ('Validate components on creation using their JSON schema '
                 '(Decreases performance).')},
    ('jsonschema', 'validate_messages'): {
        'env': 'YGG_VALIDATE_MESSAGES', 'arg': 'validate-messages',
        'type': str, 'choices': ['False', 'True', 'First'],
        'help': ('Which messages should be validated during communication. '
                 '\'True\': all messages (decreases performance), '
                 '\'False\': no messages, or '
                 '\'First\': only the first message a comm sends/receives.')},
    ('rmq', 'namespace'): {
        'env': 'YGG_NAMESPACE', 'help': 'RabbitMQ namespace.'},
    ('rmq', 'host'): {
        'env': 'YGG_MSG_HOST', 'help': 'RabbitMQ host address.'},
    ('rmq', 'vhost'): {
        'env': 'YGG_MSG_VHOST', 'help': 'RabbitMQ virtual host address.'},
    ('rmq', 'user'): {
        'env': 'YGG_MSG_USER', 'help': 'RabbitMQ username.'},
    ('rmq', 'password'): {
        'env': 'YGG_MSG_PW', 'help': 'RabbitMQ password.'},
    ('parallel', 'cluster'): {
        'env': 'YGG_CLUSTER', 'help': 'Cluster that should be used.'},
    ('testing', 'enable_examples'): {
        'env': 'YGG_ENABLE_EXAMPLE_TESTS', 'arg': 'with-examples',
        'action': 'store_true',
        'help': 'Run example tests when encountered.'},
    ('testing', 'enable_demos'): {
        'env': 'YGG_ENABLE_DEMO_TESTS', 'arg': 'with-demos',
        'action': 'store_true',
        'help': 'Run demo tests when encountered.'},
    ('testing', 'languages'): {
        'env': 'YGG_TEST_LANGUAGE', 'arg': 'language',
        'default': [], 'nargs': '+', 'type': str,
        'help': 'Language(s) that sould not be tested.'},
    ('testing', 'skip_languages'): {
        'env': 'YGG_TEST_SKIP_LANGUAGE', 'arg': 'skip-language',
        'default': [], 'nargs': '+', 'type': str,
        'help': 'Language(s) that sould not be tested.'},
    ('testing', 'enable_long'): {
        'env': 'YGG_ENABLE_LONG_TESTS', 'arg': 'long-running',
        'action': 'store_true',
        'help': 'Run long tests when encounterd.'},
    ('testing', 'enable_production_runs'): {
        'env': 'YGG_TEST_PRODUCTION_RUNS',
        'action': 'store_true',
        'help': 'Run production level tests when encountered.'},
    ('general', 'default_comm'): {
        'env': 'YGG_DEFAULT_COMM', 'type': str,
        'help': 'Comm type that should be used by default.'}
}
_key2env = {}
for k, v in _cfg_map.items():
    arg = []
    if 'arg' in v:
        arg = [v['arg']]
    if k[0] not in ['debug']:
        new_arg = k[1].replace('_', '-')
        if new_arg not in arg:
            arg.append(new_arg)
    for x in copy.deepcopy(arg):
        _key2env[x.replace('-', '_')] = v['env']
        new_arg = x.replace('-', '')
        if new_arg not in arg:
            arg.append(new_arg)
    v['args'] = arg
    v['kwargs'] = {
        kk: v[kk] for kk in ['default', 'type', 'nargs', 'help', 'choices',
                             'action']
        if kk in v}

    
class YggConfigParser(configparser.ConfigParser, object):
    r"""Config parser that returns None if option not provided on get."""

    def __init__(self, files=None):
        self.files = files
        super(YggConfigParser, self).__init__()

    def reload(self):
        r"""Reload parameters from the original files."""
        for s in self.sections():
            self.remove_section(s)
        self._sections = self._dict()
        if self.files is not None:
            self.read(self.files)

    @property
    def file_to_update(self):
        r"""str: Full path to file that should be updated if update_file is
        called without an explicit file path."""
        out = None
        if self.files is not None:
            out = self.files[-1]
        return out

    def update_file(self, fname=None):
        r"""Write out updated contents to a file.

        Args:
            fname (str, optional): Full path to file where contents should be
               saved. If None, file_to_update is used. Defaults to None.

        Raises:
            RuntimeError: If fname is None and file_to_update is None.

        """
        if fname is None:
            fname = self.file_to_update
        if fname is None:
            raise RuntimeError("No file provided or set at creation.")
        with open(fname, 'w') as fd:
            self.write(fd)

    def read(self, *args, **kwargs):
        out = super(YggConfigParser, self).read(*args, **kwargs)
        alias_map = [(('debug', 'psi'), ('debug', 'ygg')),
                     (('debug', 'cis'), ('debug', 'ygg'))]
        for old, new in alias_map:
            v = self.get(*old)
            if v:  # pragma: debug
                self.set(new[0], new[1], v)
        return out

    @classmethod
    def from_files(cls, files, **kwargs):
        r"""Construct a config parser from a set of files.

        Args:
            files (list): One or more files that options should be read from in
                the order they should be loaded.
            **kwargs: Additional keyword arguments are passed to the class
                constructor.

        Returns:
           YggConfigParser: Config parser with information loaded from the
               provided files.

        """
        out = cls(files=files, **kwargs)
        out.reload()
        return out

    def set(self, section, option, value=None):
        """Set an option."""
        if not isinstance(value, str):
            value = json.dumps(value)
        super(YggConfigParser, self).set(section, option, value=value)

    def backwards_str2val(self, val):  # pragma: no cover
        try:
            out = json.loads(val)
        except ValueError:
            if val.startswith('[') and val.endswith(']'):
                if val[1:-1]:
                    out = [self.backwards_str2val(x.strip())
                           for x in val[1:-1].split(',')]
                else:
                    out = []
            elif val.startswith("'") and val.endswith("'"):
                out = val.strip("'")
            else:
                out = val
        return out

    def get(self, section, option, default=None, **kwargs):
        r"""Return None if the section/option does not exist.

        Args:
            section (str): Name of section.
            option (str): Name of option in section.
            default (obj, optional): Value that should be returned if the
                section and/or option are not found or are an empty string.
                Defaults to None.
            **kwargs: Additional keyword arguments are passed to the parent
                class's get.

        Returns:
            obj: String entry if the section & option exist, otherwise default.

        """
        section = section.lower()
        option = option.lower()
        out = None
        if (out is None) and ((section, option) in _cfg_map):
            out = os.environ.get(_cfg_map[(section, option)]['env'], None)
        if (((out is None) and self.has_section(section)
             and self.has_option(section, option))):
            # Super does not work for ConfigParser as not inherited from object
            out = configparser.ConfigParser.get(self, section, option, **kwargs)
        # Count empty strings as not provided
        if out:
            return self.backwards_str2val(out)
        return default


def get_language_order(drivers):
    r"""Get the correct language order, including any base languages.

    Args:
        drivers (dict): Drivers in order.

    Returns:
        dict: Drivers in sorted order.

    """
    from yggdrasil.components import import_component
    out = OrderedDict()
    for d, drv in drivers.items():
        if d == 'cpp':
            d = 'c++'
        new_deps = OrderedDict()
        sub_deps = OrderedDict()
        for sub_d in drv.base_languages:
            if sub_d in drivers:
                sub_deps[sub_d] = drivers[sub_d]
            else:
                sub_deps[sub_d] = import_component('model', sub_d)
        sub_deps = get_language_order(sub_deps)
        min_dep = -1
        for sub_d, sub_drv in sub_deps.items():
            if sub_d in out:
                min_dep = max(min_dep, list(out.keys()).index(sub_d))
            else:
                new_deps[sub_d] = sub_drv
        if d in out.keys():
            dpos = list(out.keys()).index(d)
            assert(dpos > min_dep)
            min_dep = dpos
        else:
            new_deps[d] = drv
        new_out = OrderedDict()
        for k in list(out.keys())[:(min_dep + 1)]:
            new_out[k] = out[k]
        new_out.update(new_deps)
        for k in list(out.keys())[(min_dep + 1):]:
            new_out[k] = out[k]
        out = new_out
    return out


def update_language_config(languages=None, skip_warnings=False,
                           disable_languages=None, enable_languages=None,
                           allow_multiple_omp=None, lang_kwargs=None,
                           overwrite=False, verbose=False):
    r"""Update configuration options for a language driver.

    Args:
        languages (list, optional): List of languages to configure.
            Defaults to None and all supported languages will be
            configured.
        skip_warnings (bool, optional): If True, warnings about missing options
            will not be raised. Defaults to False.
        disable_languages (list, optional): List of languages that should be
            disabled. Defaults to an empty list.
        enable_languages (list, optional): List of languages that should be
            enabled. Defaults to an empty list.
        allow_multiple_omp (bool, optional): Set the allow_multiple_omp config
            option controlling whether or not the KMP_DUPLICATE_LIB_OK environment
            variable is set for model environments. Defaults to None and is
            ignored.
        overwrite (bool, optional): If True, the existing file will be overwritten.
            Defaults to False.
        verbose (bool, optional): If True, information about the config file
            will be displayed. Defaults to False.
        lang_kwargs (dict, optional): Dictionary containing language
            specific keyword arguments. Defaults to {}.

    """
    from yggdrasil.components import import_component
    if verbose:
        logger.info("Updating user configuration file for yggdrasil at:\n\t%s"
                    % usr_config_file)
    miss = []
    if (languages is None) or overwrite:
        all_languages = tools.get_supported_lang()
        languages = ['c', 'c++', 'make', 'cmake', 'python', 'lpy', 'r', 'matlab']
        for lang in all_languages:
            if lang.lower() not in languages:
                languages.append(lang)
    elif not isinstance(languages, list):
        languages = [languages]
    if disable_languages is None:
        disable_languages = []
    if enable_languages is None:
        enable_languages = []
    if lang_kwargs is None:
        lang_kwargs = {}
    if overwrite:
        shutil.copy(def_config_file, usr_config_file)
        ygg_cfg_usr.reload()
    if allow_multiple_omp is not None:
        if not ygg_cfg_usr.has_section('general'):
            ygg_cfg_usr.add_section('general')
        ygg_cfg_usr.set('general', 'allow_multiple_omp', allow_multiple_omp)
    drivers = OrderedDict([(lang, import_component('model', lang))
                           for lang in languages])
    drv = list(get_language_order(drivers).values())
    for idrv in drv:
        if (((idrv.language in disable_languages)
             and (idrv.language in enable_languages))):
            logger.info(("%s language both enabled and disabled. "
                         "No action will be taken.") % idrv.language)
        elif idrv.language in disable_languages:
            if not ygg_cfg_usr.has_section(idrv.language):
                ygg_cfg_usr.add_section(idrv.language)
            ygg_cfg_usr.set(idrv.language, 'disable', 'True')
        elif idrv.language in enable_languages:
            if not ygg_cfg_usr.has_section(idrv.language):
                ygg_cfg_usr.add_section(idrv.language)
            ygg_cfg_usr.set(idrv.language, 'disable', 'False')
        if ygg_cfg_usr.get(idrv.language, 'disable', 'False').lower() == 'true':
            continue  # pragma: no cover
        miss += idrv.configure(ygg_cfg_usr,
                               **lang_kwargs.get(idrv.language, {}))
    ygg_cfg_usr.update_file()
    ygg_cfg.reload()
    if not skip_warnings:
        for sect, opt, desc in miss:  # pragma: windows
            warnings.warn(("Could not set option %s in section %s. "
                           + "Please set this in %s to: %s")
                          % (opt, sect, ygg_cfg_usr.file_to_update, desc),
                          RuntimeWarning)
    if verbose:
        with open(usr_config_file, 'r') as fd:
            print(fd.read())
    

def get_ygg_loglevel(cfg=None, default='DEBUG'):
    r"""Get the current log level.

    Args:
        cfg (:class:`yggdrasil.config.YggConfigParser`, optional):
            Config parser with options that should be used to determine the
            log level. Defaults to :data:`yggdrasil.config.ygg_cfg`.
        default (str, optional): Log level that should be returned if the log
            level option is not set in cfg. Defaults to 'DEBUG'.

    Returns:
        str: Log level string.

    """
    is_model = tools.is_subprocess()
    if cfg is None:
        cfg = ygg_cfg
    if is_model:
        opt = 'client'
    else:
        opt = 'ygg'
    return cfg.get('debug', opt, default)


def set_ygg_loglevel(level, cfg=None):
    r"""Set the current log level.

    Args:
        level (str): Level that the log should be set to.
        cfg (:class:`yggdrasil.config.YggConfigParser`, optional):
            Config parser with options that should be used to update the
            environment. Defaults to :data:`yggdrasil.config.ygg_cfg`.
    
    """
    is_model = tools.is_subprocess()
    if cfg is None:
        cfg = ygg_cfg
    if is_model:
        opt = 'client'
    else:
        opt = 'ygg'
    cfg.set('debug', opt, level)
    logLevelYGG = eval('logging.%s' % level)
    ygg_logger = logging.getLogger("yggdrasil")
    ygg_logger.setLevel(level=logLevelYGG)
        
    
def cfg_logging(cfg=None):
    r"""Set logging levels from config options.

    Args:
        cfg (:class:`yggdrasil.config.YggConfigParser`, optional):
            Config parser with options that should be used to update the
            environment. Defaults to :data:`yggdrasil.config.ygg_cfg`.

    """
    is_model = tools.is_subprocess()
    to_stdout = is_model
    try:  # pragma: no cover
        # Direct log messages to stdout in interpreter so that messages
        # are not red in notebooks
        get_ipython  # noqa: F821
        in_notebook = True
    except BaseException:
        in_notebook = False
    to_stdout = (is_model or in_notebook)
    if cfg is None:
        cfg = ygg_cfg
    log_format = cfg.get(
        'debug', 'format',
        "%(levelname)s:%(process)d:%(module)s.%(funcName)s[%(lineno)d]:%(message)s")
    logLevelYGG = eval(
        'logging.%s' % os.environ.get(
            'YGG_DEBUG', cfg.get('debug', 'ygg', 'NOTSET')))
    logLevelRMQ = eval(
        'logging.%s' % os.environ.get(
            'YGG_RMQ_DEBUG', cfg.get('debug', 'rmq', 'INFO')))
    logLevelCLI = eval(
        'logging.%s' % os.environ.get(
            'YGG_CLIENT_DEBUG', cfg.get('debug', 'client', 'INFO')))
    if is_model:
        logLevelYGG = os.environ.get('YGG_MODEL_DEBUG', logLevelCLI)
    if not to_stdout:
        logging.basicConfig(format=log_format)
    ygg_logger = logging.getLogger("yggdrasil")
    rmq_logger = logging.getLogger("pika")
    ygg_logger.setLevel(level=logLevelYGG)
    rmq_logger.setLevel(level=logLevelRMQ)
    # For models, route the logs to stdout so that they are
    # displayed by the model driver.
    if to_stdout:
        formatter = logging.Formatter(fmt=log_format)
        handler = logging.StreamHandler(sys.stdout)
        handler.setFormatter(formatter)
        handler.setLevel(logLevelYGG)
        ygg_logger.handlers = [handler]
        rmq_logger.handlers = [handler]
        if in_notebook:  # pragma: no cover
            ygg_logger.propagate = False
            rmq_logger.propagate = False
        # ygg_logger.addHandler(handler)
        # rmq_logger.addHandler(handler)


def cfg_environment(env=None, cfg=None):
    r"""Set environment variables based on config options.

    Args:
        env (dict, optional): Dictionary of environment variables that should
            be updated. Defaults to `os.environ`.
        cfg (:class:`yggdrasil.config.YggConfigParser`, optional):
            Config parser with options that should be used to update the
            environment. Defaults to :data:`yggdrasil.config.ygg_cfg`.

    """
    if env is None:
        env = os.environ
    if cfg is None:
        cfg = ygg_cfg
    for k, v in _cfg_map.items():
        val = cfg.get(*k)
        if val:
            env.setdefault(v['env'], val)

            
# Initialize config
ygg_cfg_usr = YggConfigParser.from_files([usr_config_file])
ygg_cfg = YggConfigParser.from_files([def_config_file, usr_config_file,
                                      loc_config_file])


# Do initial update of logging & environment (legacy)
cfg_logging()
cfg_environment()


def get_config_parser(parser=None, description=None, skip_sections=None):
    r"""Create a parser or add to an existing parser arguments based on
    configuration options.

    Args:
        parser (argparse.ArgumentParser, optional): Existing argument parser
            that arguments should be added to. Defaults to None and a new
            parser is created.
        description (str, optional): The description that should be used if
            a new parser is created. Defaults to None.
        skip_sections (list, optional): Configuration sections that should be
            skipped when adding arguments to the parser. Defaults to None and
            arguments for all configuration sections will be added.

    Returns:
        argparse.ArgumentParser: Argument parser with arguments added that
            are associated with configuration options.

    """
    if skip_sections is None:
        skip_sections = []
    if parser is None:
        parser = argparse.ArgumentParser(description=description)
    for k, v in _cfg_map.items():
        if k[0] in skip_sections:
            continue
        parser.add_argument(*['--' + x for x in v['args']], **v['kwargs'])
    parser.add_argument('--production-run', action='store_true',
                        help=('Turn off safe guards in order to improve '
                              'performance. This is equivalent to '
                              '\'--validate-components '
                              '--validate-messages=True\''))
    parser.add_argument('--debug', action='store_true',
                        help=('Turn on debugging utilties including '
                              'increased logging and validation. This is '
                              'equivalent to \'--loglevel=DEBUG '
                              '--client-loglevel=DEBUG '
                              '--validate-components=False '
                              '--validate-messages=False\''))
    return parser


def resolve_config_parser(args):
    r"""Process argument results by setting flags that are set by combination
    arguments.

    Args:
        args (argparse.ArgumentParser): Argument parser to set combination
            child flags for.

    Returns:
        argparse.ArgumentParser: Argument parser with child flags for
            combination flags set.

    """
    if args.debug and args.production_run:  # pragma: debug
        raise ValueError("\'--debug\' and \'--production-run\' flags are "
                         "incompatible.")
    if args.production_run:
        args.validate_components = False
        args.validate_messages = False
    elif args.debug:
        args.loglevel = 'DEBUG'
        args.client_loglevel = 'DEBUG'
        args.validate_components = True
        args.validate_messages = True
    else:
        if args.loglevel is None:
            args.loglevel = 'INFO'
        if args.client_loglevel is None:
            args.client_loglevel = 'INFO'
    if args.validate_messages in ['True', 'False']:
        args.validate_messages = (args.validate_messages == 'True')
    return args


def acquire_env(new_env):
    r"""Get the existing environment variable values and set the environment
    based on the provided dictionary.

    Args:
        new_env (dict): Mapping from configuration key to values that
            environment variables should be set to.

    Returns:
        dict: Mapping of environment variables and values that were overridden
            by new_env.

    """
    old_env = {}
    if new_env.get('debug', '') and new_env.get('production_run', ''):  # pragma: debug
        raise ValueError("'debug' and 'production_run' variables are "
                         "incompatible.")
    if 'production_run' in new_env:
        if new_env['production_run']:
            new_env['validate_components'] = False
            new_env['validate_messages'] = False
        new_env.pop('production_run')
    if 'debug' in new_env:
        if new_env['debug']:
            new_env['loglevel'] = 'DEBUG'
            new_env['client_loglevel'] = 'DEBUG'
            new_env['validate_components'] = True
            new_env['validate_messages'] = True
    if new_env.get('validate_messages', '') in ['True', 'False']:
        new_env['validate_messages'] = (new_env['validate_messages'] == 'True')
    # old_env = {k: os.environ.get(k, None) for k in _key2env.values()}
    for k, v in new_env.items():
        k_env = _key2env.get(k, k)
        old_env.setdefault(k_env, os.environ.get(k_env, None))
        if (((k_env in ['YGG_TEST_LANGUAGE', 'YGG_TEST_SKIP_LANGUAGE'])
             and isinstance(v, list))):
            if len(v) == 0:
                continue
            from yggdrasil.components import import_component
            v = [import_component('model', x).language for x in v]
        if not isinstance(v, str):
            v = json.dumps(v)
        os.environ[k_env] = v
    if new_env.get('loglevel', False):
        set_ygg_loglevel(new_env['loglevel'])
    return old_env


def restore_env(old_env):
    r"""Restore environment variables to a previous state.

    Args:
        old_env (dict): Mapping from environment variable to value for state
            that should be restored.

    """
    for k, v in old_env.items():
        if v is None:
            os.environ.pop(k, None)
        else:
            os.environ[k] = v


@contextmanager
def parser_config(args, **kwargs):
    r"""Context manager for a run that modifies configuration options using
    values from an argument parser.

    Args:
        args (argparse.Namespace): Argument parsing results.
        **kwargs: Additional environment variable key/value pairs and/or
            argument name key/value pairs that should be added to the
            environment.

    """
    args = resolve_config_parser(args)
    for k0 in _key2env.keys():
        k = k0.replace('-', '_')
        if getattr(args, k, None) is not None:
            kwargs[k0] = getattr(args, k)
    old_env = acquire_env(kwargs)
    try:
        yield
    finally:
        restore_env(old_env)


@contextmanager
def temp_config(**kwargs):
    r"""Context manager for a run that modifies configuration options using
    a dictionary of key/value pairs.

    Args:
        **kwargs: Environment variable key/value pairs and/or argument name
            key/value pairs that should be added to the environment.

    """
    old_env = acquire_env(kwargs)
    try:
        yield
    finally:
        restore_env(old_env)



from yggdrasil.drivers.ConnectionDriver import ConnectionDriver


class RPCResponseDriver(ConnectionDriver):
    r"""Class for handling client side RPC type communication.

    Args:
        model_response_address (str): The address of the channel used by the
            client model to receive responses.
        msg_id (str): ID associate with the request message this driver was
            created to respond to.
        **kwargs: Additional keyword arguments are passed to parent class.

    Attributes:
        msg_id (str): ID associate with the request message this driver was
            created to respond to.
        response_drivers (list): Response drivers created for each request.

    """

    _connection_type = 'rpc_response'

    def __init__(self, model_response_address, msg_id, **kwargs):
        # Input communicator
        inputs = kwargs.get('inputs', [{}])
        inputs[0]['name'] = 'server_model_response.' + msg_id
        inputs[0]['is_response_server'] = True
        kwargs['inputs'] = inputs
        # Output communicator
        outputs = kwargs.get('outputs', [{}])
        outputs[0]['name'] = 'client_model_response.' + msg_id
        outputs[0]['is_response_client'] = True
        if model_response_address is not None:
            outputs[0]['address'] = model_response_address
        kwargs['outputs'] = outputs
        # Overall keywords
        super(RPCResponseDriver, self).__init__('rpc_response.' + msg_id,
                                                **kwargs)
        self.msg_id = msg_id

    @property
    def response_address(self):
        r"""str: Address of response comm."""
        return self.icomm.address

    def send_eof(self):
        r"""Send EOF message.

        Returns:
            bool: Success or failure of send.

        """
        # Don't send EOF
        return False
        
    def send_message(self, msg, **kwargs):
        r"""Propagate the request_id.

        Args:
            msg (CommMessage): Message being sent.
            **kwargs: Keyword arguments are passed to parent class send_message.

        Returns:
            bool: Success or failure of send.

        """
        if msg.header and ('request_id' in msg.header):
            kwargs.setdefault('header_kwargs', {})
            kwargs['header_kwargs'].setdefault(
                'request_id', msg.header['request_id'])
        return super(RPCResponseDriver, self).send_message(msg, **kwargs)



import copy
from yggdrasil.drivers import create_driver
from yggdrasil.drivers.Driver import Driver
        

class DuplicatedModelDriver(Driver):
    r"""Base class for Model drivers and for running executable based models.

    Args:
        name (str): Unique name used to identify the model. This will
            be used to report errors associated with the model.
        *args: Additional arguments are passed to the models in the set.
        **kwargs: Additional keyword arguments are passed to the models in
            the set.

    Attributes:

    Raises:
        RuntimeError: If both with_strace and with_valgrind are True.

    """

    name_format = "%s_copy%d"

    def __init__(self, yml, **kwargs):
        kwargs.update(yml)
        self.copies = []
        for iyml in self.get_yaml_copies(yml):
            ikws = copy.deepcopy(kwargs)
            ikws.update(iyml)
            self.copies.append(create_driver(yml=iyml, **ikws))
        super(DuplicatedModelDriver, self).__init__(**kwargs)

    @classmethod
    def get_yaml_copies(cls, yml):
        r"""Get a list of yamls for creating duplicate models for the model
        described by the provided yaml.

        Args:
            yml (dict): Input parameters for creating a model driver.

        Returns:
            list: Copies of input parameters for creating duplicate models.

        """
        env_copy_specific = {}
        for i in range(yml['copies']):
            iname = cls.name_format % (yml['name'], i)
            env_copy_specific[iname] = yml.pop('env_%s' % iname, {})
        copies = []
        for i in range(yml['copies']):
            iyml = copy.deepcopy(yml)
            iyml['name'] = cls.name_format % (yml['name'], i)
            iyml['copy_index'] = i
            # Update environment to reflect addition of suffix
            iyml['env'] = yml.get('env', {}).copy()
            iyml['env'].update(env_copy_specific.get(iyml['name'], {}))
            copies.append(iyml)
        return copies

    def cleanup(self, *args, **kwargs):
        r"""Actions to perform to clean up the thread after it has stopped."""
        for x in self.copies:
            x.cleanup(*args, **kwargs)
        super(DuplicatedModelDriver, self).cleanup(*args, **kwargs)

    def start(self, *args, **kwargs):
        r"""Start thread/process and print info."""
        # self.delay_start(*args, **kwargs)
        for x in self.copies:
            x.start(*args, **kwargs)
        super(DuplicatedModelDriver, self).start(*args, **kwargs)

    # def delay_start(self, *args, **kwargs):
    #     r"""This method should not be called in production and is only
    #     used for local testing to simulation a delayed start for some
    #     copies."""
    #     self.copies[0].start(*args, **kwargs)
    #     def start_remainder():
    #         for x in self.copies[1:]:
    #             x.start(*args, **kwargs)
    #     self.sched_task(0.4, start_remainder)
    #     super(DuplicatedModelDriver, self).start(*args, **kwargs)

    def stop(self, *args, **kwargs):
        r"""Stop the driver."""
        for x in self.copies:
            x.stop(*args, **kwargs)
        super(DuplicatedModelDriver, self).stop(*args, **kwargs)
        
    def graceful_stop(self, *args, **kwargs):
        r"""Gracefully stop the driver."""
        for x in self.copies:
            x.graceful_stop(*args, **kwargs)
        super(DuplicatedModelDriver, self).graceful_stop(*args, **kwargs)
        
    def terminate(self, *args, **kwargs):
        r"""Set the terminate event and wait for the thread/process to stop."""
        for x in self.copies:
            x.terminate(*args, **kwargs)
        super(DuplicatedModelDriver, self).terminate(*args, **kwargs)
        
    def run_loop(self):
        r"""Loop to check if model is still running and forward output."""
        # TODO: Stop if there is an error on one?
        if any([x.is_alive() for x in self.copies]):
            self.sleep()
            return
        else:
            self.set_break_flag()

    def after_loop(self, *args, **kwargs):
        r"""Actions to perform after run_loop has finished."""
        for x in self.copies:
            x.terminate()
        super(DuplicatedModelDriver, self).after_loop(*args, **kwargs)

    def printStatus(self, *args, **kwargs):
        r"""Print the class status."""
        for x in self.copies:
            x.printStatus(*args, **kwargs)
        super(DuplicatedModelDriver, self).printStatus(*args, **kwargs)

    @property
    def io_errors(self):
        r"""list: Errors produced by input/output drivers to this model."""
        errors = []
        for x in self.copies:
            errors += x.io_errors
        return errors
        
    @property
    def errors(self):
        r"""list: Errors returned by model copies."""
        out = []
        for x in self.copies:
            out += x.errors
        return out

    @errors.setter
    def errors(self, val):
        pass



import pprint
import collections
import numpy as np
import copy
from yggdrasil.tests import YggTestClass
from yggdrasil.communication import new_comm


class TestTransformBase(YggTestClass):
    r"""Test for TransformBase communication flass."""

    transform = 'TransformBase'
    skip_comm_check = True
    
    @property
    def cls(self):
        r"""str: Communication class."""
        return self.transform

    @property
    def mod(self):
        r"""str: Absolute module import."""
        return 'yggdrasil.communication.transforms.%s' % self.cls

    @property
    def inst_kwargs(self):
        r"""dict: Keyword arguments for tested class."""
        out = {}
        opt = self.get_options()
        if opt:
            out = opt[0].get('kwargs', {})
        return out

    def assert_equal_msg(self, a, b, **kwargs):
        r"""Assert that two message are the same.

        Args:
            a (object): Message for comarison.
            b (object): Expected message.
            original (object, None): Original message.

        """
        try:
            if ((isinstance(a, collections.abc.Iterator)
                 and isinstance(b, collections.abc.Iterator))):
                self.assert_equal(list(copy.deepcopy(a)), list(copy.deepcopy(b)))
            else:
                self.assert_equal(a, b)
        except BaseException:  # pragma: debug
            labels = ['expected', 'out']
            values = [b, a]
            if 'original' in kwargs:
                labels.insert(0, 'in')
                values.insert(0, kwargs['original'])
            for t, x in zip(labels, values):
                print('%s:' % t)
                if isinstance(x, collections.abc.Iterator):
                    print('iter(%s)' % pprint.pformat(list(x)))
                else:
                    pprint.pprint(x)
            raise

    def test_transform(self):
        r"""Test transform."""
        for x in self.get_options():
            inst = self.import_cls(**x.get('kwargs', {}))
            for msg_in, msg_exp in x.get('in/out', []):
                if isinstance(msg_in, collections.abc.Iterator):
                    msg_in0 = copy.deepcopy(msg_in)
                else:
                    msg_in0 = msg_in
                if isinstance(msg_exp, type(BaseException)):
                    self.assert_raises(msg_exp, inst, msg_in)
                else:
                    msg_out = inst(msg_in)
                    self.assert_equal_msg(msg_out, msg_exp, original=msg_in0)

    def test_transform_type(self):
        r"""Test transform_type."""
        for x in self.get_options():
            inst = self.import_cls(**x.get('kwargs', {}))
            for typ_in, typ_exp in x.get('in/out_t', []):
                if isinstance(typ_exp, type(BaseException)):
                    self.assert_raises(typ_exp, inst.validate_datatype,
                                       typ_in)
                else:
                    inst.validate_datatype(typ_in)
                    typ_out = inst.transform_datatype(typ_in)
                    self.assert_equal_msg(typ_out, typ_exp, original=typ_in)

    def test_transform_empty(self):
        r"""Test transform of empty bytes message."""
        self.assert_equal(self.instance(b'', no_init=True), b'')

    def test_send_comm(self):
        r"""Test transform within send comm."""
        if self.transform in ['pandas', 'PandasTransform',
                              'array', 'ArrayTransform']:
            # These transformation are negated on send
            return
        for x in self.get_options():
            if (((not x.get('in/out', False))
                 or isinstance(x['in/out'][0][1],
                               type(BaseException)))):  # pragma: debug
                continue
            msg_in, msg_out = x['in/out'][0]
            send_comm = new_comm('test_send', reverse_names=True,
                                 direction='send', use_async=False,
                                 transform=[self.import_cls(**x.get('kwargs', {}))])
            recv_comm = new_comm('test_recv', **send_comm.opp_comm_kwargs())
            if isinstance(msg_in, collections.abc.Iterator):
                msg_out_list = list(msg_out)
                msg_out = iter(msg_out_list)
            else:
                msg_out_list = [msg_out]
            try:
                for imsg_out in msg_out_list:
                    flag = send_comm.send(msg_in, timeout=self.timeout)
                    assert(flag)
                    if self.transform in ['iterate', 'IterateTransform']:
                        for iimsg_out in imsg_out:
                            flag, msg_recv = recv_comm.recv(timeout=self.timeout)
                            assert(flag)
                            self.assert_equal_msg(msg_recv, iimsg_out,
                                                  original=msg_in)
                        msg_recv = imsg_out
                    elif ((self.transform in ['filter', 'FilterTransform'])
                          and isinstance(imsg_out, collections.abc.Iterator)):
                        assert(recv_comm.n_msg_recv == 0)
                        flag, msg_recv = recv_comm.recv(timeout=0.0)
                        assert(flag)
                        self.assert_equal_msg(msg_recv, b'', original=msg_in)
                        msg_recv = imsg_out
                    elif ((self.transform in ['map', 'MapFieldsTransform',
                                              'select_fields', 'SelectFieldsTransform'])
                          and isinstance(imsg_out, np.ndarray)):
                        flag, msg_recv = recv_comm.recv_array(timeout=self.timeout)
                    else:
                        flag, msg_recv = recv_comm.recv(timeout=self.timeout)
                    assert(flag)
                    self.assert_equal_msg(msg_recv, imsg_out,
                                          original=msg_in)
            finally:
                send_comm.close()
                recv_comm.close()

    def test_recv_comm(self):
        r"""Test transform within recv comm."""
        for x in self.get_options():
            if (((not x.get('in/out', False))
                 or isinstance(x['in/out'][0][1],
                               type(BaseException)))):  # pragma: debug
                continue
            msg_in, msg_out = x['in/out'][0]
            send_comm = new_comm('test_send', reverse_names=True,
                                 direction='send', use_async=False)
            recv_comm = new_comm('test_recv',
                                 transform=[self.import_cls(**x.get('kwargs', {}))],
                                 **send_comm.opp_comm_kwargs())
            if isinstance(msg_in, collections.abc.Iterator):
                msg_out_list = list(msg_out)
                msg_out = iter(msg_out_list)
            else:
                msg_out_list = [msg_out]
            try:
                flag = send_comm.send(msg_in, timeout=self.timeout)
                assert(flag)
                for imsg_out in msg_out_list:
                    if (((self.transform in ['map', 'MapFieldsTransform',
                                             'select_fields', 'SelectFieldsTransform'])
                         and isinstance(imsg_out, np.ndarray))):
                        flag, msg_recv = recv_comm.recv_array(timeout=self.timeout)
                    else:
                        flag, msg_recv = recv_comm.recv(timeout=self.timeout)
                    assert(flag)
                    self.assert_equal_msg(msg_recv, imsg_out, original=msg_in)
            finally:
                send_comm.close()
                recv_comm.close()



from yggdrasil.communication import FileComm


class WOFOSTParamFileComm(FileComm.FileComm):
    r"""Class for handling I/O from/to a WOFOST parameter file on disk."""

    _filetype = 'wofost'
    _schema_subtype_description = 'The file is a WOFOST parameter file.'
    _default_serializer = 'wofost'



import os
import sys
import six
import atexit
import weakref
import logging
import threading
import queue
import multiprocessing
from yggdrasil.tools import YggClass


mp_ctx = multiprocessing.get_context()
mp_ctx_spawn = multiprocessing.get_context("spawn")
_main_thread = threading.main_thread()
_thread_registry = weakref.WeakValueDictionary()
_lock_registry = weakref.WeakValueDictionary()
logger = logging.getLogger(__name__)


def check_processes():  # pragma: debug
    r"""Check for processes that are still running."""
    import psutil
    current_process = psutil.Process()
    children = current_process.children(recursive=True)
    if len(children) > 0:
        logging.info("Process %s has %d children" % (
            current_process.pid, len(children)))
        for child in children:
            logger.info("    %s process running" % child.pid)


def check_threads():  # pragma: debug
    r"""Check for threads that are still running."""
    # logger.info("Checking %d threads" % len(_thread_registry))
    for k, v in _thread_registry.items():
        if v.is_alive():
            logger.error("Thread is alive: %s" % k)
    if threading.active_count() > 1:
        logger.info("%d threads running" % threading.active_count())
        for t in threading.enumerate():
            logger.info("    %s thread running" % t.name)


def check_locks():  # pragma: debug
    r"""Check for locks in lock registry that are locked."""
    # logger.info("Checking %d locks" % len(_lock_registry))
    for k, v in _lock_registry.items():
        res = v.acquire(False)
        if res:
            v.release()
        else:
            logger.error("Lock could not be acquired: %s" % k)


def check_sockets():  # pragma: debug
    r"""Check registered sockets."""
    from yggdrasil.communication import cleanup_comms
    count = cleanup_comms('ZMQComm')
    if count > 0:
        logger.info("%d sockets closed." % count)


def ygg_atexit():  # pragma: debug
    r"""Things to do at exit."""
    check_locks()
    check_threads()
    # # This causes a segfault in a C dependency
    # if not is_subprocess():
    #     check_sockets()
    # Python 3.4 no longer supported if using pip 9.0.0, but this
    # allows the code to work if somehow installed using an older
    # version of pip
    if sys.version_info[0:2] == (3, 4):  # pragma: no cover
        # Print empty line to ensure close
        print('', end='')
        sys.stdout.flush()


atexit.register(ygg_atexit)


class SafeThread(threading.Thread):
    r"""Thread that sets Event on error."""

    def __init__(self, *args, **kwargs):
        self._errored = threading.Event()
        super(SafeThread, self).__init__(*args, **kwargs)

    def run(self, *args, **kwargs):
        try:
            super(SafeThread, self).run(*args, **kwargs)
        except BaseException:
            self._errored.set()
            raise

    @property
    def exitcode(self):
        r"""int: Exit code. 1 if error, 0 otherwise."""
        if (not self._started.is_set()) or self.is_alive():
            return None
        return int(self._errored.is_set())

    @property
    def pid(self):
        r"""Process ID."""
        return os.getpid()


class AliasDisconnectError(RuntimeError):
    pass


def add_aliased_attribute(cls, name, with_lock=False):
    r"""Factory to alias an attribute so that it refers to the wrapped
    object.

    Args:
        name (str): Name of attribute to alias.
        with_lock (bool, optional): If True, the class's lock will be
            acquired before getting the attribute. Defaults to False.

    """
    def alias_wrapper(self):
        self.check_for_base(name)
        lock_acquired = False
        if ((with_lock and hasattr(self, 'lock')
             and (name not in self._unlocked_attr))):
            self.lock.acquire()
            lock_acquired = True
        try:
            out = getattr(self._base, name)
        finally:
            if lock_acquired:
                self.lock.release()
        return out
    alias_wrapper.__name__ = name
    setattr(cls, name, property(alias_wrapper))


def add_aliased_method(cls, name, with_lock=False):
    r"""Factory to alias a method so that it refers to the wrapped
    object.

    Args:
        name (str): Name of method to alias.
        with_lock (bool, optional): If True, the class's lock will be
            acquired before executing the method. Defaults to False.

    """
    def alias_wrapper(self, *args, **kwargs):
        self.check_for_base(name)
        lock_acquired = False
        if ((with_lock and hasattr(self, 'lock')
             and (name not in self._unlocked_attr))):
            self.lock.acquire()
            lock_acquired = True
        try:
            out = getattr(self._base, name)(*args, **kwargs)
        finally:
            if lock_acquired:
                self.lock.release()
        return out
    alias_wrapper.__name__ = name
    setattr(cls, name, alias_wrapper)


class AliasMeta(type):
    r"""Meta class for adding aliased methods to the class."""

    def __new__(meta, name, bases, class_dict):
        cls = type.__new__(meta, name, bases, class_dict)
        for k in cls._base_meth:
            assert(not hasattr(cls, k))
            add_aliased_method(cls, k, with_lock=cls._base_locked)
        for k in cls._base_attr:
            assert(not hasattr(cls, k))
            add_aliased_attribute(cls, k, with_lock=cls._base_locked)
        cls._base_meth = []
        cls._base_attr = []
        return cls


@six.add_metaclass(AliasMeta)
class AliasObject(object):
    r"""Alias object that calls to attribute.

    Args:
        dont_initialize_base (bool, optional): If True the base object
            will not be initialized. Defaults to False.

    """

    _base_class = None
    _base_attr = []
    _base_meth = []
    _base_locked = False
    _unlocked_attr = []

    def __init__(self, *args, dont_initialize_base=False, **kwargs):
        self._base = None
        if (not dont_initialize_base) and (self._base_class is not None):
            self._base = self._base_class(*args, **kwargs)

    @classmethod
    def from_base(cls, base, *args, **kwargs):
        r"""Create an instance by creating a based from the provided
        base class."""
        if base is not None:
            kwargs['dont_initialize_base'] = True
            out = cls(*args, **kwargs)
            out._base = base
        else:
            out = cls(*args, **kwargs)
        return out

    def __getstate__(self):
        out = self.__dict__.copy()
        out.pop('_base_class', None)
        return out

    def __setstate__(self, state):
        self.__dict__.update(state)

    def check_for_base(self, attr):
        r"""Raise an error if the aliased object has been disconnected."""
        if self._base is None:
            raise AliasDisconnectError(
                ("Aliased object has been disconnected so "
                 "'%s' is no longer available.") % attr)

    @property
    def dummy_copy(self):
        r"""Dummy copy of base."""
        return None

    def disconnect(self):
        r"""Disconnect from the aliased object by replacing it with
        a dummy object."""
        if self._base is not None:
            self._base = self.dummy_copy

    def __del__(self):
        self.disconnect()

        
class MultiObject(AliasObject):
    r"""Concurrent/parallel processing object using either threads
    or processes."""

    def __init__(self, *args, task_method="threading", **kwargs):
        self.task_method = task_method
        if task_method in ["thread", "threading", "concurrent"]:
            self.parallel = False
        elif task_method in ["process", "multiprocessing", "parallel"]:
            self.parallel = True
        else:  # pragma: debug
            raise ValueError(("Unsupported method for concurrency/"
                              "parallelism: '%s'") % task_method)
        super(MultiObject, self).__init__(*args, **kwargs)


class Context(MultiObject):
    r"""Context for managing threads/processes."""

    def __init__(self, task_method='thread', dont_initialize_base=False):
        super(Context, self).__init__(dont_initialize_base=True,
                                      task_method=task_method)
        if not dont_initialize_base:
            if self.parallel:
                self._base = mp_ctx_spawn
            else:
                self._base = threading

    def __getstate__(self):
        state = super(Context, self).__getstate__()
        if self.parallel:
            state['_base'] = state['_base']._name
        else:
            state['_base'] = None
        return state

    def __setstate__(self, state):
        if state['_base'] is None:
            state['_base'] = threading
        else:
            # Use the existing context?
            # state['_base'] = mp_ctx_spawn
            state['_base'] = multiprocessing.get_context(state['_base'])
        super(Context, self).__setstate__(state)

    def RLock(self, *args, **kwargs):
        r"""Get a recursive lock in this context."""
        kwargs['task_context'] = self
        return RLock(*args, **kwargs)

    def Event(self, *args, **kwargs):
        r"""Get an event in this context."""
        kwargs['task_context'] = self
        return Event(*args, **kwargs)

    def Task(self, *args, **kwargs):
        r"""Get a task in this context."""
        kwargs['task_context'] = self
        return Task(*args, **kwargs)

    def Queue(self, *args, **kwargs):
        r"""Get a queue in this context."""
        kwargs['task_context'] = self
        return Queue(*args, **kwargs)

    def Dict(self, *args, **kwargs):
        r"""Get a shared dictionary in this context."""
        kwargs['task_context'] = self
        return Dict(*args, **kwargs)

    def current_task(self):
        r"""Current task (process/thread)."""
        if self.parallel:
            return self._base.current_process()
        else:
            return self._base.current_thread()

    def main_task(self):
        r"""Main task (process/thread)."""
        if self.parallel:
            out = None
            if hasattr(self._base, 'parent_process'):  # pragma: no cover
                out = self._base.parent_process()
            if out is None:
                out = self.current_task()
            return out
        else:
            return _main_thread


class DummyContextObject(object):  # pragma: no cover

    @property
    def context(self):
        return None

    def disconnect(self):
        pass


class ContextObject(MultiObject):
    r"""Base class for object intialized in a context."""

    def __init__(self, *args, task_method='threading',
                 task_context=None, **kwargs):
        self._managed_context = None
        if task_context is None:
            task_context = Context(task_method=task_method)
            self._managed_context = task_context
        task_method = task_context.task_method
        self._context = weakref.ref(task_context)
        self._base_class = self.get_base_class(task_context)
        super(ContextObject, self).__init__(
            *args, task_method=task_method, **kwargs)

    def __getstate__(self):
        state = super(ContextObject, self).__getstate__()
        state['_context'] = None
        return state

    def __setstate__(self, state):
        if state['_managed_context'] is None:
            state['_managed_context'] = Context(task_method=state['task_method'])
        state['_context'] = weakref.ref(state['_managed_context'])
        super(ContextObject, self).__setstate__(state)
        
    @classmethod
    def get_base_class(cls, context):
        r"""Get instance of base class that will be represented."""
        name = cls.__name__
        context.check_for_base(name)
        return getattr(context._base, name)

    def disconnect(self):
        r"""Disconnect from the aliased object by replacing it with
        a dummy object."""
        if ContextObject is not None:
            super(ContextObject, self).disconnect()
        if self._managed_context is not None:
            self._managed_context.disconnect()
            self._managed_context = None

    @property
    def context(self):
        r"""Context: Context used to create this object."""
        return self._context()


class DummyRLock(DummyContextObject):  # pragma: no cover

    def acquire(self, *args, **kwargs):
        pass

    def release(self, *args, **kwargs):
        pass

    def __enter__(self, *args, **kwargs):
        return self

    def __exit__(self, *args, **kwargs):
        pass
    

class RLock(ContextObject):
    r"""Recursive lock. Acquiring the lock after disconnect is called
    through use as a context will not raise an error, but will not
    do anything."""

    _base_meth = ['acquire', 'release', '__enter__', '__exit__']

    def __getstate__(self):
        state = super(RLock, self).__getstate__()
        if (not self.parallel) and (not isinstance(state['_base'], DummyRLock)):
            state['_base'] = None
        return state

    def __setstate__(self, state):
        if state['_base'] is None:
            state['_base'] = threading.RLock()
        super(RLock, self).__setstate__(state)

    @property
    def dummy_copy(self):
        r"""Dummy copy of base."""
        return DummyRLock()


class DummyEvent(DummyContextObject):  # pragma: no cover

    def __init__(self, value=False):
        self._value = value

    def is_set(self):
        return self._value

    def set(self):
        self._value = True

    def clear(self):
        self._value = False

    def wait(self, *args, **kwargs):
        if self._value:
            return
        raise AliasDisconnectError("DummyEvent will never change to True.")


class Event(ContextObject):
    r"""Multiprocessing/threading event."""

    _base_attr = (ContextObject._base_attr
                  + ['is_set', 'set', 'clear', 'wait'])

    @property
    def dummy_copy(self):
        r"""Dummy copy of base."""
        return DummyEvent(self._base.is_set())
        
    def __getstate__(self):
        state = super(Event, self).__getstate__()
        if not self.parallel:
            state['_base'] = state['_base'].is_set()
        return state

    def __setstate__(self, state):
        if isinstance(state['_base'], bool):
            val = state['_base']
            state['_base'] = threading.Event()
            if val:
                state['_base'].set()
        super(Event, self).__setstate__(state)


class DummyTask(DummyContextObject):  # pragma: no cover

    def __init__(self, name='', exitcode=0, daemon=False):
        self.name = name
        self.exitcode = exitcode
        self.daemon = daemon
        super(DummyTask, self).__init__()

    def join(self, *args, **kwargs):
        return

    def is_alive(self):
        return False

    def terminate(self):
        pass

    def kill(self):
        pass
    

class Task(ContextObject):
    r"""Multiprocessing/threading process."""

    _base_attr = ['name', 'daemon', 'authkey', 'sentinel', 'exitcode',
                  'pid']
    _base_meth = ['start', 'run', 'join',
                  # Thread only
                  'getName', 'setName', 'isDaemon', 'setDaemon',
                  # Process only
                  'terminate']

    @classmethod
    def get_base_class(cls, context):
        r"""Get instance of base class that will be represented."""
        if context.parallel:
            return context._base.Process
        else:
            return SafeThread

    @property
    def dummy_copy(self):
        r"""Dummy copy of base."""
        name = b'dummy'
        exitcode = 0
        daemon = False
        try:
            name = self._base.name
            exitcode = self._base.exitcode
            daemon = self._base.daemon
        except AttributeError:  # pragma: debug
            pass
        return DummyTask(name=name, exitcode=exitcode, daemon=daemon)
        
    def __getstate__(self):
        state = super(Task, self).__getstate__()
        if not self.parallel:
            state['_base'] = {
                'name': state['_base'].name, 'group': None,
                'daemon': state['_base'].daemon,
                'target': state['_base']._target,
                'args': state['_base']._args,
                'kwargs': state['_base']._kwargs}
        return state

    def __setstate__(self, state):
        if isinstance(state['_base'], dict):
            state['_base'] = SafeThread(**state['_base'])
        super(Task, self).__setstate__(state)

    def is_alive(self):
        r"""Determine if the process/thread is alive."""
        out = self._base.is_alive()
        if out is None:  # pragma: debug
            out = False
        return out

    @property
    def ident(self):
        r"""Process ID."""
        if self.parallel:
            return self._base.pid
        else:
            return self._base.ident

    def kill(self, *args, **kwargs):
        r"""Kill the task."""
        if self.parallel and hasattr(self._base, 'kill'):
            return self._base.kill(*args, **kwargs)
        elif hasattr(self._base, 'terminate'):
            return self._base.terminate(*args, **kwargs)


class DummyQueue(DummyContextObject):  # pragma: no cover

    def empty(self):
        return True

    def full(self):
        return False

    def get(self, *args, **kwargs):
        raise AliasDisconnectError("There are no messages in a DummyQueue.")

    def get_nowait(self, *args, **kwargs):
        raise AliasDisconnectError("There are no messages in a DummyQueue.")

    def put(self, *args, **kwargs):
        raise AliasDisconnectError("Cannot put messages in a DummyQueue.")

    def put_nowait(self, *args, **kwargs):
        raise AliasDisconnectError("Cannot put messages in a DummyQueue.")

    def qsize(self):
        return 0

    def join(self, *args, **kwargs):
        return

    def join_thread(self, *args, **kwargs):
        return
    
    def close(self):
        pass


class Queue(ContextObject):
    r"""Multiprocessing/threading queue."""

    _base_meth = ['full', 'get', 'get_nowait', 'join_thread', 'qsize']
    
    @classmethod
    def get_base_class(cls, context):
        r"""Get instance of base class that will be represented."""
        if context.parallel:
            return context._base.Queue
        else:
            return queue.Queue

    def __getstate__(self):
        state = super(Queue, self).__getstate__()
        if (not self.parallel) and (not isinstance(state['_base'], DummyQueue)):
            state['_base'] = None
        return state

    def __setstate__(self, state):
        if state['_base'] is None:
            state['_base'] = queue.Queue()
        super(Queue, self).__setstate__(state)

    @property
    def dummy_copy(self):
        r"""Dummy copy of base."""
        return DummyQueue()

    def join(self, *args, **kwargs):
        self.check_for_base('join')
        if self.parallel:
            try:
                self._base.close()
            except OSError:  # pragma: debug
                pass
            return self._base.join_thread(*args, **kwargs)
        else:
            return self._base.join(*args, **kwargs)

    def disconnect(self):
        r"""Disconnect from the aliased object by replacing it with
        a dummy object."""
        if self.parallel:
            self.join()
        if Queue is not None:
            super(Queue, self).disconnect()

    def empty(self):
        try:
            return self._base.empty()
        except OSError:  # pragma: debug
            self.disconnect()
            return True

    def put(self, *args, **kwargs):
        try:
            self._base.put(*args, **kwargs)
        except AttributeError:  # pragma: debug
            # Multiprocessing queue asserts it is not closed
            self.disconnect()
            raise AliasDisconnectError("Queue was closed.")

    def put_nowait(self, *args, **kwargs):
        try:
            self._base.put_nowait(*args, **kwargs)
        except AttributeError:  # pragma: debug
            # Multiprocessing queue asserts it is not closed
            self.disconnect()
            raise AliasDisconnectError("Queue was closed.")


class Dict(ContextObject):
    r"""Multiprocessing/threading shared dictionary."""

    _base_meth = ['clear', 'copy', 'get', 'items', 'keys',
                  'pop', 'popitem', 'setdefault', 'update', 'values',
                  '__contains__', '__delitem__', '__getitem__',
                  '__iter__', '__len__', '__setitem__']
    
    @classmethod
    def get_base_class(cls, context):
        r"""Get instance of base class that will be represented."""
        if context.parallel:
            manager = context._base.Manager()
            return manager.dict
        else:
            return LockedDict

    # Don't define this so that is is not called after manager is
    # shut down.
    # @property
    # def dummy_copy(self):
    #     r"""Dummy copy of base."""
    #     return self._base.copy()
        
    def disconnect(self):
        r"""Disconnect from the aliased object by replacing it with
        a dummy object."""
        try:
            final_value = self._base.copy()
        except BaseException:  # pragma: debug
            final_value = {}
        if isinstance(self._base, LockedDict):
            self._base.disconnect()
        if getattr(self._base, '_manager', None) is not None:
            self._base._manager.shutdown()
        if hasattr(self._base, '_close'):
            self._base._close()
        if Dict is not None:
            super(Dict, self).disconnect()
        self._base = final_value


class LockedObject(AliasObject):
    r"""Container that provides a lock that is acquired before accessing
    the object."""

    _base_locked = True
    
    def __init__(self, *args, task_method='process',
                 task_context=None, **kwargs):
        self.lock = RLock(task_method=task_method,
                          task_context=task_context)
        super(LockedObject, self).__init__(*args, **kwargs)

    def disconnect(self):
        r"""Disconnect from the aliased object by replacing it with
        a dummy object."""
        if LockedObject is not None:
            super(LockedObject, self).disconnect()
        self.lock.disconnect()


# class LockedList(LockedObject):
#     r"""List intended to be shared between threads."""

#     def __init__(self, *args, **kwargs):
#         base = list(*args, **kwargs)
#         super(LockedList, self).__init__(base)


class LockedDict(LockedObject):
    r"""Dictionary that can be shared between threads."""

    _base_class = dict
    _base_meth = ['clear', 'copy', 'get', 'items', 'keys',
                  'pop', 'popitem', 'setdefault', 'update', 'values',
                  '__contains__', '__delitem__', '__getitem__',
                  '__iter__', '__len__', '__setitem__']
    
    def add_subdict(self, key):
        r"""Add a subdictionary."""
        self[key] = {}

    @property
    def dummy_copy(self):
        r"""Dummy copy of base."""
        try:
            out = self._base.copy()
        except BaseException:  # pragma: debug
            out = {}
        return out


# class LockedWeakValueDict(LockedDict):
#     r"""Dictionary of weakrefs that can be shared between threads."""

#     _base_class = weakref.WeakValueDictionary
#     _base_attr = ['data']
#     _base_meth = ['itervaluerefs', 'valuerefs']
    
#     def __init__(self, *args, **kwargs):
#         self._dict_refs = {}
#         super(LockedWeakValueDict, self).__init__(*args, **kwargs)

#     def add_subdict(self, key):
#         r"""Add a subdictionary."""
#         self._dict_refs[key] = weakref.WeakValueDictionary()
#         self[key] = self._dict_refs[key]


class YggTask(YggClass):
    r"""Class for managing Ygg thread/process."""

    _disconnect_attr = (YggClass._disconnect_attr
                        + ['context', 'lock', 'process_instance',
                           'error_flag', 'start_flag', 'terminate_flag'])
    
    def __init__(self, name=None, target=None, args=(), kwargs=None,
                 daemon=False, group=None, task_method='thread',
                 context=None, with_pipe=False, **ygg_kwargs):
        if kwargs is None:
            kwargs = {}
        if (target is not None) and ('target' in self._schema_properties):
            ygg_kwargs['target'] = target
            target = None
        self.context = Context.from_base(task_method=task_method,
                                         base=context)
        self.as_process = self.context.parallel
        if self.as_process:
            self.in_process = False
            self.pipe = None
            self.send_pipe = None
            if with_pipe:
                self.pipe = self.context._base.Pipe()
                kwargs['send_pipe'] = self.pipe[1]
        else:
            self.in_process = True
        process_kwargs = dict(
            name=name, group=group, daemon=daemon,
            target=self.run)
        self.process_instance = self.context.Task(**process_kwargs)
        self._ygg_target = target
        self._ygg_args = args
        self._ygg_kwargs = kwargs
        self.lock = self.context.RLock()
        self.create_flag_attr('error_flag')
        self.create_flag_attr('start_flag')
        self.create_flag_attr('terminate_flag')
        self._calling_thread = None
        self.state = ''
        super(YggTask, self).__init__(name, **ygg_kwargs)
        if not self.as_process:
            global _thread_registry
            global _lock_registry
            _thread_registry[self.name] = self.process_instance._base
            _lock_registry[self.name] = self.lock._base
            atexit.register(self.atexit)

    def __getstate__(self):
        out = super(YggTask, self).__getstate__()
        out.pop('_input_args', None)
        out.pop('_input_kwargs', None)
        return out

    def atexit(self):  # pragma: debug
        r"""Actions performed when python exits."""
        if self.is_alive():
            self.info('Thread alive at exit')
            self.cleanup()

    def printStatus(self):
        r"""Print the class status."""
        self.logger.info('%s(%s): state: %s', self.__module__,
                         self.print_name, self.state)

    def cleanup(self):
        r"""Actions to perform to clean up the thread after it has stopped."""
        self.disconnect()
        
    def create_flag_attr(self, attr):
        r"""Create a flag."""
        setattr(self, attr, self.context.Event())

    def get_flag_attr(self, attr):
        r"""Return the flag attribute."""
        return getattr(self, attr)

    def set_flag_attr(self, attr, value=True):
        r"""Set a flag."""
        if value:
            self.get_flag_attr(attr).set()
        else:
            self.get_flag_attr(attr).clear()

    def clear_flag_attr(self, attr):
        r"""Clear a flag."""
        self.set_flag_attr(attr, value=False)

    def check_flag_attr(self, attr):
        r"""Determine if a flag is set."""
        return self.get_flag_attr(attr).is_set()

    def wait_flag_attr(self, attr, timeout=None):
        r"""Wait until a flag is True."""
        return self.get_flag_attr(attr).wait(timeout=timeout)

    def start(self, *args, **kwargs):
        r"""Start thread/process and print info."""
        self.state = 'starting'
        if not self.was_terminated:
            self.set_started_flag()
            self.before_start()
        self.process_instance.start(*args, **kwargs)
        # self._calling_thread = self.get_current_task()

    def before_start(self):
        r"""Actions to perform on the main thread/process before
        starting the thread/process."""
        self.debug('')

    def run(self, *args, **kwargs):
        r"""Continue running until terminate event set."""
        self.debug("Starting method")
        self.state = 'running'
        try:
            self.run_init()
            self.call_target()
        except BaseException:  # pragma: debug
            self.state = 'error'
            self.run_error()
        finally:
            self.run_finally()
            if self.state != 'error':
                self.state = 'finished'

    def run_init(self):
        r"""Actions to perform at beginning of run."""
        # atexit.register(self.atexit)
        self.debug('pid = %s, ident = %s', self.pid, self.ident)
        self.in_process = True
        if self.as_process and ('send_pipe' in self._ygg_kwargs):
            self.send_pipe = self._ygg_kwargs.pop('send_pipe')

    def call_target(self):
        r"""Call target."""
        if self._ygg_target:
            self._ygg_target(*self._ygg_args, **self._ygg_kwargs)

    def run_error(self):
        r"""Actions to perform on error in try/except wrapping run."""
        self.exception("%s ERROR", self.context.task_method.upper())
        self.set_flag_attr('error_flag')

    def run_finally(self):
        r"""Actions to perform in finally clause of try/except wrapping
        run."""
        if self.as_process:
            if self.send_pipe is not None:
                self.send_pipe.close()
        for k in ['_ygg_target', '_ygg_args', '_ygg_kwargs']:
            if hasattr(self, k):
                delattr(self, k)

    def join(self, *args, **kwargs):
        r"""Join the process/thread."""
        return self.process_instance.join(*args, **kwargs)

    def is_alive(self, *args, **kwargs):
        r"""Determine if the process/thread is alive."""
        return self.process_instance.is_alive(*args, **kwargs)

    @property
    def pid(self):
        r"""Process ID."""
        return self.process_instance.pid

    @property
    def ident(self):
        r"""Process ID."""
        return self.process_instance.ident

    @property
    def daemon(self):
        r"""bool: Indicates whether the thread/process is daemonic or not."""
        return self.process_instance.daemon
        
    @property
    def exitcode(self):
        r"""Exit code."""
        if self.as_process:
            out = int(self.check_flag_attr('error_flag'))
            if self.process_instance.exitcode:
                out = self.process_instance.exitcode
            return out
        else:
            return int(self.check_flag_attr('error_flag'))

    @property
    def returncode(self):
        r"""Return code."""
        return self.exitcode

    def kill(self, *args, **kwargs):
        r"""Kill the process."""
        self.process_instance.kill(*args, **kwargs)
        return self.terminate(*args, **kwargs)

    def terminate(self, no_wait=False):
        r"""Set the terminate event and wait for the thread/process to stop.

        Args:
            no_wait (bool, optional): If True, terminate will not block until
                the thread/process stops. Defaults to False and blocks.

        Raises:
            AssertionError: If no_wait is False and the thread/process has not
                stopped after the timeout.

        """
        self.debug('')
        with self.lock:
            self.state = 'terminated'
            if self.was_terminated:  # pragma: debug
                self.debug('Driver already terminated.')
                return
            self.set_terminated_flag()
        if not no_wait:
            # if self.is_alive():
            #     self.join(self.timeout)
            self.wait(timeout=self.timeout)
            assert(not self.is_alive())
        # if self.as_process:
        #     self.process_instance.terminate()

    def poll(self):
        r"""Check if the process is finished and return the return
        code if it is."""
        out = None
        if not self.is_alive():
            out = self.returncode
        return out

    def get_current_task(self):
        r"""Get the current process/thread."""
        return self.context.current_task()

    def get_main_proc(self):
        r"""Get the main process/thread."""
        return self.context.main_task()

    def set_started_flag(self, value=True):
        r"""Set the started flag for the thread/process to True."""
        self.set_flag_attr('start_flag', value=value)

    def set_terminated_flag(self, value=True):
        r"""Set the terminated flag for the thread/process to True."""
        self.set_flag_attr('terminate_flag', value=value)

    @property
    def was_started(self):
        r"""bool: True if the thread/process was started. False otherwise."""
        return self.check_flag_attr('start_flag')

    @property
    def was_terminated(self):
        r"""bool: True if the thread/process was terminated. False otherwise."""
        return self.check_flag_attr('terminate_flag')

    @property
    def main_terminated(self):
        r"""bool: True if the main thread/process has terminated."""
        return (not self.get_main_proc().is_alive())

    def wait(self, timeout=None, key=None):
        r"""Wait until thread/process finish to return using sleeps rather than
        blocking.

        Args:
            timeout (float, optional): Maximum time that should be waited for
                the driver to finish. Defaults to None and is infinite.
            key (str, optional): Key that should be used to register the timeout.
                Defaults to None and is set based on the stack trace.

        """
        T = self.start_timeout(timeout, key_level=1, key=key)
        while self.is_alive() and not T.is_out:
            self.verbose_debug('Waiting for %s to finish...',
                               self.context.task_method)
            self.sleep()
        self.stop_timeout(key_level=1, key=key)


class BreakLoopException(BaseException):
    r"""Special exception that can be raised by the target function
    for a loop in order to break the loop."""

    def __init__(self, *args, **kwargs):
        import traceback
        self.break_stack = ''.join(traceback.format_stack())
        super(BreakLoopException, self).__init__(*args, **kwargs)


class BreakLoopError(BreakLoopException):
    r"""Version of BreakLoopException that sets an error message."""
    pass
        

class YggTaskLoop(YggTask):
    r"""Class to run a loop inside a thread/process."""

    _disconnect_attr = (YggTask._disconnect_attr
                        + ['break_flag', 'loop_flag'])

    def __init__(self, *args, **kwargs):
        super(YggTaskLoop, self).__init__(*args, **kwargs)
        self._1st_main_terminated = False
        self._loop_count = 0
        self.create_flag_attr('break_flag')
        self.create_flag_attr('loop_flag')
        self.create_flag_attr('unpause_flag')
        self.set_flag_attr('unpause_flag', value=True)
        self.break_stack = None

    @property
    def loop_count(self):
        r"""int: Number of loops performed."""
        with self.lock:
            return self._loop_count

    def on_main_terminated(self, dont_break=False):  # pragma: debug
        r"""Actions performed when 1st main terminated.

        Args:
            dont_break (bool, optional): If True, the break flag won't be set.
                Defaults to False.

        """
        self._1st_main_terminated = True
        if not dont_break:
            self.debug("on_main_terminated")
            self.set_break_flag()

    def set_break_flag(self, value=True, break_stack=None):
        r"""Set the break flag for the thread/process to True."""
        if self.break_stack is None:
            if break_stack is None:
                import traceback
                break_stack = ''.join(traceback.format_stack())
            self.break_stack = break_stack
        self.set_flag_attr('break_flag', value=value)
        if value:
            self.set_flag_attr('unpause_flag', value=True)

    def pause(self):
        r"""Pause the loop execution."""
        self.set_flag_attr('unpause_flag', value=False)

    def resume(self):
        r"""Resume the loop execution."""
        self.set_flag_attr('unpause_flag', value=True)

    @property
    def was_break(self):
        r"""bool: True if the break flag was set."""
        return self.check_flag_attr('break_flag')

    def set_loop_flag(self, value=True):
        r"""Set the loop flag for the thread/process to True."""
        self.set_flag_attr('loop_flag', value=value)

    @property
    def was_loop(self):
        r"""bool: True if the thread/process was loop. False otherwise."""
        return self.check_flag_attr('loop_flag')

    def wait_for_loop(self, timeout=None, key=None, nloop=0):
        r"""Wait until thread/process enters loop to return using sleeps rather than
        blocking.

        Args:
            timeout (float, optional): Maximum time that should be waited for
                the thread/process to enter loop. Defaults to None and is infinite.
            key (str, optional): Key that should be used to register the timeout.
                Defaults to None and is set based on the stack trace.
            nloop (int, optional): Number of loops that should be performed
                before breaking. Defaults to 0.

        """
        T = self.start_timeout(timeout, key_level=1, key=key)
        while (((not self.was_loop) or (self.loop_count < nloop))
               and self.is_alive() and (not T.is_out)):  # pragma: debug
            self.verbose_debug('Waiting for thread/process to enter loop...')
            self.sleep()
        self.stop_timeout(key_level=1, key=key)

    def before_loop(self):
        r"""Actions performed before the loop."""
        self.debug('')

    def after_loop(self):
        r"""Actions performed after the loop."""
        self.debug('')

    def call_target(self):
        r"""Call target."""
        self.debug("Starting loop")
        self.before_loop()
        if (not self.was_break):
            self.set_loop_flag()
        while (not self.was_break):
            if ((self.main_terminated
                 and (not self._1st_main_terminated))):  # pragma: debug
                self.on_main_terminated()
            else:
                self.wait_flag_attr('unpause_flag')
                try:
                    self.run_loop()
                except BreakLoopError as e:
                    self.error("BreakLoopError: %s", e)
                    self.set_break_flag(break_stack=e.break_stack)
                except BreakLoopException as e:
                    self.debug("BreakLoopException: %s", e)
                    self.set_break_flag(break_stack=e.break_stack)
        if not self.break_stack:
            self.set_break_flag()
        
    def run_loop(self, *args, **kwargs):
        r"""Actions performed on each loop iteration."""
        if self._ygg_target:
            self._ygg_target(*self._ygg_args, **self._ygg_kwargs)
        else:
            self.set_break_flag()
        with self.lock:
            self._loop_count += 1

    def run_error(self):
        r"""Actions to perform on error in try/except wrapping run."""
        super(YggTaskLoop, self).run_error()
        self.debug("run_error")
        self.set_break_flag()
        
    def run(self, *args, **kwargs):
        r"""Continue running until terminate event set."""
        super(YggTaskLoop, self).run(*args, **kwargs)
        try:
            self.after_loop()
        except BaseException:  # pragma: debug
            self.exception("AFTER LOOP ERROR")
            self.set_flag_attr('error_flag')

    def terminate(self, *args, **kwargs):
        r"""Also set break flag."""
        self.debug("terminate")
        self.set_break_flag()
        super(YggTaskLoop, self).terminate(*args, **kwargs)



import os
import argparse


def run(tmin, tmax, tstep):
    t = tmin
    mass = 0

    while t <= tmax:

        # Calculate mass for the time step
        # (pretend this is a biologically complex calculation)
        mass += t * 0.2

        # Advance time step
        t += tstep

    return mass


# Parse command-line arguments
if __name__ == "__main__":
    parser = argparse.ArgumentParser("Simulate root growth over time.")
    parser.add_argument('tmin', help='Starting time (in days)', type=float)
    parser.add_argument('tmax', help='Ending time (in days)', type=float)
    parser.add_argument('tstep', help='Time step (in days)', type=float)
    args = parser.parse_args()
    run(args.tmin, args.tmax, args.tstep)
